import traceback
import networkx as nx
from .CFG import CFGGraph
from ...utils import java_nodes
from loguru import logger

class CFGGraph_java(CFGGraph):
    def __init__(self, src_language, src_code, properties, root_node, parser):
        super().__init__(src_language, src_code, properties, root_node, parser)

        self.node_list = None
        self.statement_types = java_nodes.statement_types
        self.CFG_node_list = []
        self.CFG_edge_list = []
        self.records = {
            "basic_blocks": {},
            "method_list": {},
            "constructor_list": {},
            "return_type": {},
            "class_list": {},
            "extends": {},
            "function_calls": {},
            "constructor_calls": {},
            "object_instantiate": {},
            "switch_child_map": {},
            "label_statement_map": {},
            "return_statement_map": {},
            "lambda_map": {},
        }
        self.types = ["scoped_type_identifier", "type_identifier", "generic_type"]
        self.index_counter = max(self.index.values())
        self.CFG_node_indices = []
        self.symbol_table = self.parser.symbol_table
        self.declaration = self.parser.declaration
        self.declaration_map = self.parser.declaration_map
        self.CFG_node_list, self.CFG_edge_list = self.CFG_java()
        self.graph = self.to_networkx(self.CFG_node_list, self.CFG_edge_list)
        

    def get_index(self, node):
        return self.index[(node.start_point, node.end_point, node.type)]
    
    def check_inner_class(self, node):
        while node.parent is not None:
            if node.parent.type == "class_declaration":
                return True
            node = node.parent
        return False
    
    def get_basic_blocks(self, CFG_node_list, CFG_edge_list):
        G = self.to_networkx(CFG_node_list, CFG_edge_list)
        components = nx.weakly_connected_components(G)
        # NOTE: May need to sort these components according to line number (this one is more foolproof but harder to implement) or the first element in the set (less foolproof, but I think it can be proved, easier to implement).
        # As of now coincidentally the AST node numbering and the list of nodes in networkx are according ot the line numbers
        block_index = 1
        for block in components:
            block_list = sorted(list(block))
            self.records["basic_blocks"][block_index] = block_list
            block_index += 1

    def get_key(self, val, dictionary):
        for key, value in dictionary.items():
            if val in value:
                return key
    def get_containing_method(self, node):
        while node is not None:
            if node.type == 'lambda_expression':
                return node
            if node.type == "method_declaration" or node.type == "constructor_declaration":
                return node
            if node.type == "static_initializer":
                return node
            node = node.parent
        while node is not None:
            if node.type in self.statement_types["node_list_type"]:
                return node
            node = node.parent

    def get_lambda_body(self, node):
        """Returns the body of a lambda expression, breadthfirst"""
        bfs_queue = []
        bfs_queue.append(node)
        while bfs_queue != []:
            top = bfs_queue.pop(0)
            if top.type == "lambda_expression":
                return top
            for child in top.children:
                bfs_queue.append(child)
        return None
    
    def get_all_lambda_body(self, node):
        """Returns the body of a lambda expression, breadthfirst"""
        bfs_queue = []
        output = []
        bfs_queue.append(node)
        while bfs_queue != []:
            top = bfs_queue.pop(0)
            if top.type == "lambda_expression":
                output.append(top)
                # return [top]
            for child in top.children:
                if child.type == "lambda_expression" or child.type not in self.statement_types["node_list_type"]:
                    bfs_queue.append(child)

        return output

    def append_block_index(self, CFG_node_list):
        new_list = []
        for node in CFG_node_list:
            block_index = self.get_key(node[0], self.records["basic_blocks"])
            new_list.append((node[0], node[1], node[2], node[3], block_index))
        return new_list

    def add_edge(self, src_node, dest_node, edge_type, additional_data=None):
        # current_node_list = list(map(lambda x: x[0], self.CFG_node_list))
        if src_node == None or dest_node == None:
            logger.error(
                "Node where adding edge is attempted is none {}->{}",
                src_node,
                dest_node,
            )
            logger.warning(traceback.format_stack()[-2])
            print(src_node, dest_node, edge_type)
            raise NotImplementedError
        else:
            for src, dest, ty, *_ in self.CFG_edge_list:
                if src == src_node and dest == dest_node:
                    return
            self.CFG_edge_list.append((src_node, dest_node, edge_type, additional_data))

    def handle_next(self, src_node, dest_node, edge_type):
        if dest_node == None:
            try:
                current_containing_method = self.get_index(self.get_containing_method(src_node))
            except:
                return
            try:
                self.records["return_statement_map"][current_containing_method].append(self.get_index(src_node))
            except:
                self.records["return_statement_map"][current_containing_method] = [self.get_index(src_node)]
            
        else:
            self.add_edge(self.get_index(src_node), self.get_index(dest_node), edge_type)

    def get_next_index(self, node_value):
        # If exiting a method, don't return
        next_node = node_value.next_named_sibling
        while next_node is not None and next_node.type == "block" and len(list(filter(lambda child : child.is_named, next_node.children))) == 0:
            next_node = next_node.next_named_sibling
        if next_node is not None and next_node.type == "class_declaration":
            # Check if its a local class
            if self.check_inner_class(node_value):
                next_node = next_node.next_named_sibling
        next_node_index = -1  # Just using a dummy initial value
        if next_node == None:
            current_node = node_value
            while current_node.parent is not None:
                if (current_node.parent.type in self.statement_types["loop_control_statement"]):
                    next_node = current_node.parent
                    break
                if current_node.parent.type == "if_statement":
                    next_node_index, next_node = self.get_next_index(current_node.parent)
                    break
                next_node = current_node.next_named_sibling
                if current_node.type in self.statement_types["definition_types"]:
                    next_node = None
                    break
                if next_node is not None:
                    break
                current_node = current_node.parent

        if next_node == None:
            return 2, None
        if next_node.type in self.statement_types["definition_types"]:
            return 2, None
        try:
            # WARNING: Might need to replace if with while
            if next_node.type == "block":
                for child in next_node.children:
                    if child.is_named:
                        next_node = child
                        break
            # Returns the index of the next node and the node object of the next node
            try:
                current_containing_method = self.get_index(self.get_containing_method(node_value))
                next_containing_method = self.get_index(self.get_containing_method(next_node))
            except:
                return 2, None
            if current_containing_method != next_containing_method:
                return 2, None
            else:
                return (self.get_index(next_node), next_node)
        
        except Exception as e:
            # return 2, None
            print("DO NOT IGNORE", e)
            logger.warning(traceback.format_stack()[-2])
            raise NotImplementedError
            return next_node_index, next_node

    def edge_first_line(self, current_node_key, current_node_value):
        # We need to add an edge to the first statement in the next basic block
        node_index = self.index[current_node_key]
        try:
            current_block_index = self.get_key(node_index, self.records["basic_blocks"])
            next_block_index = current_block_index + 1
            first_line_index = self.records["basic_blocks"][next_block_index][0]
            src_node = node_index
            dest_node = first_line_index
            self.add_edge(src_node, dest_node, "first_next_line")  # We could maybe differentiate this
        except:
            # Most probably the block is empty
            # add a direct edge to the next statement
            if current_node_value.type == "method_declaration":
                return
            next_index, next_node = self.get_next_index(current_node_value)
            self.handle_next(current_node_value, next_node, "next_line 9")

    def edge_to_body(self, current_node_key, current_node_value, body_type, edge_type):
        # We need to add an edge to the first statement in the body block
        src_node = self.index[current_node_key]
        body_node = current_node_value.child_by_field_name(body_type)
        flag = False
        while body_node.type == "block":
            for child in body_node.children:
                if child.is_named:
                    flag = True
                    body_node = child
                    break
            if flag == False:
                return
        if (body_node.is_named and body_node.type in self.statement_types["node_list_type"]):
            dest_node = self.get_index(body_node)
            self.add_edge(src_node, dest_node, edge_type)
        

    def get_block_last_line(self, current_node_value, block_type):
        # Find the last line in the body block
        block_node = current_node_value.child_by_field_name(block_type)
        if block_node is None:
            for child in reversed(current_node_value.children):
                if child.is_named:
                    block_node = child
                    break

        if block_node.is_named is False:
            return (current_node_value, current_node_value.type)

        while block_node.type in self.statement_types["statement_holders"]:
            named_children = list(filter(lambda child: child.is_named == True, reversed(block_node.children)))
            if len(named_children) == 0:
                # It means there is an empty block - thats why no named nodes inside
                return (current_node_value, current_node_value.type)
                    
            block_node = named_children[0]
            if block_node.type in self.statement_types["node_list_type"]:
                return (block_node, block_node.type)
        return (block_node, block_node.type)
    
    def get_class_name(self,node):
        reference_node = node.child_by_field_name("object")
        types = ["scoped_type_identifier", "type_identifier", "generic_type"]

        if reference_node is None or reference_node.type == "this":
            while node is not None:
                if node.type == "class_declaration":
                    # class_index = self.get_index(node)
                    class_name = list(filter(lambda child: child.type == "identifier", node.children))[0]
                    class_name = [class_name.text.decode("UTF-8")]
                    break
                node = node.parent
            try:
                class_name += self.records['extends'][class_name[0]]
            except:
                pass
            return class_name
        try:
            reference_index = self.get_index(reference_node)
            declaration_index = self.declaration_map[reference_index]
            class_name = [self.symbol_table["data_type"][declaration_index]]
            try:
                class_name += self.records['extends'][class_name[0]]
                
            except:
                pass
            return class_name

        except:
            # If a static method has been explicitly called on a class name
            if reference_node is not None:
                if reference_node.type == "object_creation_expression":
                    class_name = list(filter(lambda child: child.type in types, reference_node.children))[0]
                    class_name = [class_name.text.decode("UTF-8")]
                else:
                    class_name = [reference_node.text.decode("UTF-8")]
                try:
                    class_name += self.records['extends'][class_name[0]]
                except:
                    pass
                return class_name
        
            return None
    def get_return_type(self,current_node):
        method_name = current_node.child_by_field_name("name").text.decode("UTF-8")
        class_name = self.get_class_name(current_node)
        if class_name is not None:
            class_name = class_name[0]
            method_name = (class_name, method_name)
        else:
            method_name = (None, method_name)
        signature = ()
        argument_list = current_node.child_by_field_name("arguments")
        argument_list = list(filter(lambda child : child.is_named, argument_list.children))
        signature = self.get_signature(argument_list)
        function_key = (method_name, signature)
        try:
            data_type = self.records["return_type"][function_key]
        except:
            data_type = "void"
        # records["return_type"][((class_name,method_name), signature)] = return_type
        return data_type 
    
    def get_signature(self, argument_list):
        literal_type_map = {
            "character_literal": "char",
            "string_literal": "String",
            "decimal_integer_literal": "int",
            "boolean": "boolean",
            "decimal_floating_point_literal": "double", # If float, won't work
        }
        signature = []
        for argument in argument_list:
            if argument.type == "identifier":
                identifier_index = self.get_index(argument)
                try:
                    declaration_index = self.declaration_map[identifier_index]
                    data_type = self.symbol_table["data_type"][declaration_index]
                except Exception as e:
                    data_type = "Unknown"
                signature.append(data_type)
            elif argument.type == "method_invocation":
                try:
                    data_type = self.get_return_type(argument)
                except:
                    data_type = "Unknown"
                signature.append(data_type)
            elif argument.type == "field_access":
                field_variable = argument.children[-1]
                identifier_index = self.get_index(field_variable)
                try:
                    declaration_index = self.declaration_map[identifier_index]
                    data_type = self.symbol_table["data_type"][declaration_index]
                except Exception as e:
                    data_type = "Unknown"
                signature.append(data_type)
            elif argument.type == "this":
                signature.append(self.get_class_name(argument)[0])
            else:
                try:
                    signature.append(literal_type_map[argument.type])
                except:
                    signature.append("Unknown")
        return tuple(signature)
    
    def function_list(self, current_node, node_list):
        current_index = self.get_index(current_node)
        if current_node.type == "method_invocation":
            parent_node = None
            pointer_node = current_node
            while pointer_node is not None:
                if (pointer_node.parent is not None and pointer_node.parent.type in self.statement_types["node_list_type"]):
                    try:
                        p = pointer_node.parent
                        parent_node = node_list[(p.start_point, p.end_point, p.type)]
                        break
                    except Exception as e:
                        pass
                pointer_node = pointer_node.parent
            parent_index = self.get_index(parent_node)
            # maintain a list of all method invocations
            # IDENTIFY THE CLASS THAT THE ALIAS BELONGS TO AND USE THAT IN THE MAP MAYBE? 
            base_method_name = current_node.child_by_field_name("name").text.decode("UTF-8")
            signature = ()
            argument_list = current_node.child_by_field_name("arguments")
            argument_list = list(filter(lambda child : child.is_named, argument_list.children))
            signature = self.get_signature(argument_list)

            class_name_list = self.get_class_name(current_node)
            if class_name_list is None:
                method_name = (None, base_method_name)
                function_key = (method_name, signature)
                if method_name[1] != "println" and method_name[1] != "print":
                    if function_key in self.records["function_calls"].keys():
                        self.records["function_calls"][function_key].append((current_index, parent_index))
                    else:
                        self.records["function_calls"][function_key] = [(current_index,parent_index)]
            else:
                for class_name in class_name_list:
                    method_name = (class_name, base_method_name)
                    function_key = (method_name, signature)
                    if method_name[1] != "println" and method_name[1] != "print":
                        if function_key in self.records["function_calls"].keys():
                            self.records["function_calls"][function_key].append((current_index, parent_index))
                        else:
                            self.records["function_calls"][function_key] = [(current_index,parent_index)]
        elif current_node.type == "object_creation_expression":
            parent_node = None
            pointer_node = current_node
            while pointer_node is not None:
                if (pointer_node.parent is not None and pointer_node.parent.type in self.statement_types["node_list_type"]):
                    try:
                        p = pointer_node.parent
                        parent_node = node_list[(p.start_point, p.end_point, p.type)]
                        # parent_node = pointer_node.parent
                        break
                    except Exception as e:
                        pass
                pointer_node = pointer_node.parent
            parent_index = self.get_index(parent_node)
            type_name = list(filter(lambda child : child.type in self.types, current_node.children))
            type_name = type_name[0].text.decode("utf-8")
            try:
                self.records["object_instantiate"][type_name].append((current_index,parent_index))
            except:
                self.records["object_instantiate"][type_name] = [(current_index,parent_index)]
                    # break
            signature = ()
            argument_list = current_node.child_by_field_name("arguments")
            argument_list = list(filter(lambda child : child.is_named, argument_list.children))
            signature = self.get_signature(argument_list)
            method_name = (type_name, type_name)
            function_key = (method_name, signature)
            if function_key in self.records["constructor_calls"].keys():
                self.records["constructor_calls"][function_key].append((current_index, parent_index))
            else:
                self.records["constructor_calls"][function_key] = [(current_index,parent_index)]

        elif  current_node.type == "explicit_constructor_invocation":
            parent_node = current_node
            parent_index = self.get_index(parent_node)
            type_name = current_node.child_by_field_name("constructor").text.decode("utf-8")
            if type_name == 'this': # TODO: Add super also here for now
                type_name_list = self.get_class_name(current_node)
                for type_name in type_name_list:
                    # TODO: Replace this condition with a method level check flag
                    if type_name != "test":
                        try:
                            self.records["object_instantiate"][type_name].append((current_index,current_index))
                        except:
                            self.records["object_instantiate"][type_name] = [(current_index,current_index)]

                        signature = ()
                        argument_list = current_node.child_by_field_name("arguments")
                        argument_list = list(filter(lambda child : child.is_named, argument_list.children))
                        signature = self.get_signature(argument_list)
                        method_name = (type_name, type_name)
                        function_key = (method_name, signature)
                        if function_key in self.records["constructor_calls"].keys():
                            self.records["constructor_calls"][function_key].append((current_index, parent_index))
                        else:
                            self.records["constructor_calls"][function_key] = [(current_index,parent_index)]
            
            elif type_name == 'super':
                pass
            else:
                raise Exception("Explicit constructor invocation not handled")
           
        elif current_node.type == "method_declaration" or current_node.type == "constructor_declaration":
            last_line, _ = self.get_block_last_line(current_node, "body")
            try:
                self.records["return_statement_map"][current_index].append(self.get_index(last_line))
            except:
                self.records["return_statement_map"][current_index] = [self.get_index(last_line)]

        elif current_node.type == "class_declaration":
            # If constructor exits, find the last line of the constructor
            # If no constructor, find the last line before a method starts
            # If empty of if only methods, then just return from the class
            constructor_node = None
            constructor_count = 0
            empty_flag = True
            last_statement = None
            class_node = current_node.child_by_field_name("body")
            class_children = list(filter(lambda child: child.is_named and child.type != "method_declaration" and child.type != "class_declaration", class_node.children))
            for child in reversed(class_children):
                if child.type in self.statement_types["node_list_type"]:
                    empty_flag = False
                    if last_statement is None and child.type != "constructor_declaration":
                        last_statement = child
                if child.type == "constructor_declaration":
                    constructor_node = child
                    constructor_count += 1
                    break
            if empty_flag == True:
                try:
                    self.records["return_statement_map"][current_index].append(current_index)
                except:
                    self.records["return_statement_map"][current_index] = [current_index]
            elif constructor_count == 1:
                last_line, _ = self.get_block_last_line(constructor_node, "body")
                try:
                    self.records["return_statement_map"][current_index].append(self.get_index(last_line))
                except:
                    self.records["return_statement_map"][current_index] = [self.get_index(last_line)]
            elif last_statement is not None:
                last_index = self.get_index(last_statement)
                try:
                    self.records["return_statement_map"][current_index].append(last_index)
                except:
                    self.records["return_statement_map"][current_index] = [last_index]
            elif constructor_node is not None:
                last_line, _ = self.get_block_last_line(constructor_node, "body")
                try:
                    self.records["return_statement_map"][current_index].append(self.get_index(last_line))
                except:
                    self.records["return_statement_map"][current_index] = [self.get_index(last_line)]

        for child in current_node.children:
            if child.is_named:
                self.function_list(child, node_list)

    def get_all_statements(self, current_node, node_list, statements):
        for child in current_node.children:
            if child.is_named and child.type in self.statement_types["node_list_type"]:
                child_key = (child.start_point, child.end_point, child.type)
                if child_key in node_list.keys():
                    statements.append(child)
            self.get_all_statements(child, node_list, statements)

        return statements

    def add_dummy_nodes(self):
        self.CFG_node_list.append((1, 0, "start_node", "start"))
        # self.CFG_node_list.append((2, 0, "exit_node", "exit"))

    def get_signature_nodes(self, node):
        signature = []
        formal_parameters = node.child_by_field_name('parameters')
        formal_parameters = list(filter(lambda x: x.type == 'formal_parameter', formal_parameters.children))
        
        for formal_parameter in formal_parameters:
            for child in formal_parameter.children:
                if child.type != "identifier":
                    signature.append(child.text.decode('utf-8'))
        return tuple(signature)

    def get_class_name_nodes(self, node):
        type_identifiers = ["type_identifier", "generic_type", "scoped_type_identifier"]
        "Returns the class name when a method declaration or constructor declaration is passed to it"

        while node is not None:
            if node.type == "class_body" and node.parent.type == "class_declaration":
                node = node.parent
                class_index = self.index[(node.start_point, node.end_point, node.type)]
                class_name = list(filter(lambda child: child.type == "identifier", node.children))[0]
                return class_index, class_name.text.decode("UTF-8")
            elif node.type == "class_body" and node.parent.type == "object_creation_expression":
                node = node.parent
                class_index = self.index[(node.start_point, node.end_point, node.type)]
                class_name = list(filter(lambda child: child.type in type_identifiers, node.children))[0]
                return class_index, class_name.text.decode("UTF-8")
            node = node.parent
    
    def inner_function(self,current_node):
        """Returns true if the current node is inside an inner function"""
        method_counter = 0
        while current_node is not None:
            if current_node.type == "method_declaration":
                method_counter += 1
            current_node = current_node.parent
        if method_counter > 1:
            return True
        else:
            return False
    def returns_inner_definition(self, node, inner_node):
        
        """Returns the inner definition index if exists"""
        for c in node.children:
            if c.is_named and c.type == "method_declaration":
                method_name = list(filter(lambda child: child.type == "identifier", c.children))[0].text.decode("utf-8")
                _, class_name = self.get_class_name_nodes(c)
                signature = self.get_signature_nodes(c)
                # TODO: Take out the common unitility functions like get class name and signture
                inner_node.append(self.records["method_list"][((class_name,method_name), signature)])
                return
            else:
                self.returns_inner_definition(c, inner_node)
        return

    # def add_dummy_edges(self):
    #     for node_name, node_index in self.records["function_calls"].items():
    #         # node_index[0] is suppposed to be the statement node number of the function call
    #         # node_index[1] is the dummy index value assigned to the dummy node created for the external function call
    #         for node in node_index:
    #             if node_name not in self.records["method_list"].keys():
    #                 self.add_edge(node[1], node[0], "function_call")
    #                 self.add_edge(node[0], node[1], "function_return")
    #             else:
    #                 self.add_edge(
    #                     node[1],
    #                     self.records["method_list"][node_name],
    #                     "recursive_method_call",
    #                 )

    def get_matched_constructor(self, node):
        for node_signature, node_ind in self.records["constructor_calls"].items():
            for node_search in node_ind:
                if node == node_search:
                    if node_signature in self.records["constructor_list"].keys():
                        method_index = self.records["constructor_list"][node_signature]
                        return method_index
                           
        for node_signature, node_ind in self.records["constructor_calls"].items():
            for node_search in node_ind:
                flag = False
                if node == node_search:
                    for list_signature, list_index in self.records["constructor_list"].items():
                        flag = True
                        list_signature_list = list(list_signature[1])
                        node_signature_list = list(node_signature[1])
                        if len(list_signature_list) != len(node_signature_list):
                            flag = False
                        else:
                            for i in range(len(list_signature_list)):
                                if list_signature_list[i] != "Unknown" and node_signature_list[i] != "Unknown" and list_signature_list[i] != node_signature_list[i]:
                                    flag = False
                                    break
                        if flag == True:
                            return list_index


    def add_method_call_edges(self):
        # print("Method List")
        # print(*self.records["method_list"].items(), sep="\n")
        # print("Constructor List")
        # print(*self.records["constructor_list"].items(), sep="\n")
        # print("Function Calls")
        # print(*self.records["function_calls"].items(), sep="\n")
        # print("Object instantiations")
        # print(*self.records["object_instantiate"].items(), sep="\n")
        # print("Constuctor Calls")
        # print(*self.records["constructor_calls"].items(), sep="\n")
        # print("extends")
        # print(self.records["extends"])
        # print("Declaration Map")
        # print(self.declaration_map)
        # print("SymboL_TABLE.DATATYPES")
        # print(self.symbol_table["data_type"])
        # print("declaration")
        # print(self.declaration)
        # print("____________________________________________________________________________")
        # print("return_statement_map")
        # print(self.records["return_statement_map"])

        for node_signature, node_index in self.records["function_calls"].items():
            # node_index[1] is suppposed to be the statement node number of the function call
            # node_index[0] is the dummy index value assigned to the dummy node created for the external function call
            # Update: node_index[0] is no longer dummy index but the ast id of the function invocation node
            for node in node_index:
                if node_signature in self.records["method_list"].keys():
                    method_index = self.records["method_list"][node_signature]
                    # Find the class name before indexing records["method_list"]
                    edge_type = "method_call|"+str(node[0])
                    self.add_edge(node[1], self.records["method_list"][node_signature], edge_type)
                    # Add the returning edge
                    # Use the return statement index available in records. And add an edge from all the return statements to the calling line here
                    try:
                        for return_node in self.records["return_statement_map"][method_index]:
                            self.add_edge(return_node, node[1], "method_return")
                    except:
                        # If you can't find return statements, then add an edge from the last line 
                        # Handled by adding the last line to the return statement map in self.function_calls
                        pass
        for node_name, node_index in self.records["object_instantiate"].items():
            for node in node_index:
                if node_name in self.records["class_list"].keys():
                    class_index = self.records["class_list"][node_name]
                    edge_type = "constructor_call|"+str(node[0])
                    additional_data_index = self.get_matched_constructor(node)
                    additional_data = {}
                    if additional_data_index is not None:
                        additional_data = {"target_constructor": additional_data_index}
                    # for node_signature, node_ind in self.records["constructor_calls"].items():
                    #     for node_search in node_ind:
                    #         if node == node_search:
                    #             if node_signature in self.records["constructor_list"].keys():
                    #                 method_index = self.records["constructor_list"][node_signature]
                    #                 additional_data = {"target_constructor": method_index}
                    #                 print(node_signature, method_index)
                    #                 break
                    #     if additional_data:
                    #         break
                    # assert additional_data
                    # print("ADDING EDGE", node[1], class_index, edge_type, additional_data)
                    self.add_edge(node[1], class_index, edge_type, additional_data=additional_data)
                    try:
                        for return_node in self.records["return_statement_map"][class_index]:
                            self.add_edge(return_node, node[1], "class_return")
                    except Exception as e:
                        # If you can't find return statements, then add an edge from the last line 
                        # Handled by adding the last line to the return statement map in self.function_calls
                        pass

        for node_signature, node_index in self.records["constructor_calls"].items():
            for node in node_index:
                if node_signature in self.records["constructor_list"].keys():
                    method_index = self.records["constructor_list"][node_signature]
                    try:
                        for return_node in self.records["return_statement_map"][method_index]:
                            self.add_edge(return_node, node[1], "class_return")
                    except:
                        # If you can't find return statements, then add an edge from the last line 
                        # Handled by adding the last line to the return statement map in self.function_calls
                        pass
                   
        # for lambda expressions
        # for node_key, lambda_expression in self.records["lambda_map"].items():
        for lambda_key, statement_node in self.records["lambda_map"].items():
            # lambda_index = self.index[node_key]
            lambda_index = self.index[lambda_key]
            # self.edge_to_body(node_key, lambda_expression, "body", "lambda_invocation")
            self.add_edge(self.get_index(statement_node), lambda_index, "lambda_invocation")
            #TODO: maintain a return map for all points of return
            try:
                for return_node in self.records["return_statement_map"][lambda_index]:
                    self.add_edge(return_node, self.get_index(statement_node), "lambda_return 1")
            except:
                pass
                for lambda_expression in self.get_all_lambda_body(statement_node):
                # Note: Might need to move this to before the try
                    last_line, line_type = self.get_block_last_line(lambda_expression, "body")
                    if line_type not in self.statement_types['node_list_type']:
                        self.add_edge(self.get_index(lambda_expression), self.get_index(statement_node), "lambda_return 2")
                    else:
                        last_line_index = self.get_index(last_line)
                        self.add_edge(last_line_index, self.get_index(statement_node), "lambda_return 3")
            

                

    def return_next_node(self,node_value):
        # node_value = node_value.parent  
        next_node = node_value.next_named_sibling
        while next_node is None and node_value.parent.type in self.statement_types['statement_holders']:
            if node_value.parent.parent.type == 'method_declaration':         
                next_node = node_value.parent.next_named_sibling
                return next_node
            else:
                node_value = node_value.parent
                next_node = node_value.next_named_sibling   
        if node_value.parent.type in self.statement_types['statement_holders']:
            return next_node
        return None
    
    def add_class_edge(self, node_value):
        class_attributes = ["field_declaration", "static_initializer"]
        # Not included: ["record_declaration", "method_declaration","compact_constructor_declaration","class_declaration","interface_declaration","annotation_type_declaration","enum_declaration","block","static_initializer","constructor_declaration"]
        current_index = self.get_index(node_value)
        current_node = node_value.child_by_field_name("body")
        flag = False
        # Chain all class level attributes together and append to class declaration.
        current_fields = list(filter(lambda x: x.type in class_attributes, current_node.children))
        for field in current_fields:
            if field.type == "static_initializer":
                # Find the first line insdie the block
                block = list(filter(lambda x : x.type == "block", field.children))[0]
                try:
                    field = list(filter(lambda x : x.is_named, block.children))[-1]
                except:
                    continue
                field_index = self.get_index(field)    
            else:
                field_index = self.get_index(field)
                self.add_edge(current_index, field_index, "class_next")
            current_index = field_index

        # If constructor exists, chain it to the last field
        constructors = list(filter(lambda x: x.type == "constructor_declaration", current_node.children))
        for constructor in constructors:
            constructor_index = self.get_index(constructor)
            self.add_edge(current_index, constructor_index, "constructor_next")

        # In case main method exists, chain main method at the end
        try:
            methods = list(filter(lambda x: x.type == "method_declaration", current_node.children))
            for method in methods:
                if self.records["main_method"] == self.get_index(method): 
                    self.add_edge(current_index, self.records["main_method"], "main_method_next")
        except:
            pass

    def CFG_java(self):
        warning_counter = 0
        node_list = {}
        # node_list is a dictionary that maps from (node.start_point, node.end_point, node.type) to the node object of tree-sitter
        _, self.node_list, self.CFG_node_list, self.records = java_nodes.get_nodes(
            root_node=self.root_node,
            node_list=node_list,
            graph_node_list=self.CFG_node_list,
            index=self.index,
            records=self.records,
        )
        # self.CFG_node_indices = list(map(lambda x: self.index[x], node_list.keys()))
        # Initial for loop required for basic block creation and simple control flow within a block ----------------------------
        for node_key, node_value in node_list.items():
            current_node_type = node_key[2]
            if current_node_type in self.statement_types["non_control_statement"]:
                src_node = self.index[node_key]
                # if node_value.next_named_sibling is not None and node_value.next_named_sibling.type == "static_initializer":
                # if current_node_type == "field_declaration" and node_value.next_named_sibling is not None and node_value.next_named_sibling.type == "constructor_declaration":
                #     self.add_edge(src_node, self.get_index(node_value.next_named_sibling), "constructor_next 2")
                if (
                    java_nodes.return_switch_child(node_value) is None
                    and node_value.parent is not None
                    and node_value.parent.type
                    in self.statement_types["statement_holders"]
                ):
                    # There is no switch expression in the subtree starting at this statement node
                    try: 
                        check = False
                        next_node = node_value.next_named_sibling
                        if next_node is not None and next_node.type == "static_initializer":
                            try:
                                child = list(filter(lambda x : x.type == "block", next_node.children))[0]
                                next_node = child
                            except:
                                pass
                        while next_node is not None and next_node.type == "block" and len(list(filter(lambda child : child.is_named, next_node.children))) == 0:
                            next_node = next_node.next_named_sibling
                        # If it is a local class, skip it and go for the next node
                        if next_node is not None and next_node.type == "class_declaration":
                            # Check if its a local class
                            if self.check_inner_class(node_value):
                                next_node = next_node.next_named_sibling
                        # TODO: This might break next line, check all test cases
                        # if next_node is None and self.get_containing_method(node_value).type == "constructor_declaration": 
                        #     self.handle_next(node_value, None, "constructor_return")
                        if next_node is None and node_value.parent.type == 'block':
                            next_node = self.return_next_node(node_value)
                            dest_node = self.get_index(next_node)
                            check = True
                            if next_node.type in self.statement_types["node_list_type"] and next_node.type not in self.statement_types["definition_types"]:
                                self.add_edge(src_node, dest_node, "next_line $")

                        flag = False
                        while next_node.type == "block" and len(next_node.named_children):
                            for child in next_node.children:
                                if child.is_named:
                                    if child.type in self.statement_types["node_list_type"]:
                                        flag = True
                                        next_node = child
                                        break
                                    else:
                                        next_node = child
                                        break

                            if flag == True:
                                break
                        
                        
                        dest_node = self.get_index(next_node)
                        if dest_node in self.records["switch_child_map"].keys():
                            dest_node = self.records["switch_child_map"][dest_node]
                        if check is False and next_node.type in self.statement_types['node_list_type'] and next_node.type not in self.statement_types["definition_types"]:
                            self.add_edge(src_node, dest_node, "next_line 1")

                        
                    except:
                        pass

            elif current_node_type in self.statement_types["not_implemented"]:
                print("WARNING: Not implemented ", current_node_type)
                warning_counter += 1

        self.get_basic_blocks(self.CFG_node_list, self.CFG_edge_list)
        self.CFG_node_list = self.append_block_index(self.CFG_node_list)
        self.CFG_node_indices = list(map(lambda x: x[0], self.CFG_node_list))
        self.function_list(self.root_node, node_list)
        self.add_dummy_nodes()
        # self.add_dummy_edges()
        # ------------------------------------------------------------------------------
        # At this point, the self.CFG_node_list has basic block index appended to it
        # ------------------------------------------------------------------------------
        for node_key, node_value in node_list.items():
            current_node_type = node_key[2]
            current_index = self.index[node_key]
            if (current_node_type == "method_declaration" or current_node_type == "constructor_declaration"):
                try:
                    # If main method exists, this will work
                    if current_index == self.records["main_method"]:
                        self.add_edge(1, self.records["main_class"], "next")
                        # self.add_edge(self.records["main_class"], current_index, "next")
                except:
                    if node_value.parent.parent.type == "class_declaration":
                        # We need to add an edge to the first statement in the next basic block
                        self.add_edge(1, current_index, "next")

                self.edge_first_line(node_key, node_value)
                last_line, line_type = self.get_block_last_line(node_value, "body")
                last_line_index = self.get_index(last_line)
                try: 
                    if current_case_index == self.records["main_method"]:
                        if line_type in self.statement_types["non_control_statement"]:
                            if last_line_index in self.records["switch_child_map"].keys():
                                last_line_index = self.records["switch_child_map"][last_line_index]
                            # self.add_edge(last_line_index, 2, "exit_next")
                except:
                    # No main method in this snippet
                    pass
            #------------------------------------------------------------------------------
            elif current_node_type == "interface_declaration":
                self.edge_first_line(node_key, node_value)
            #------------------------------------------------------------------------------
            if current_node_type == "class_declaration":
                self.add_class_edge(node_value)
            elif current_node_type == "import_declaration":
                next_node = node_value.next_named_sibling
                if next_node is not None and next_node.type != "import_declaration":
                    try:
                        self.add_edge(current_index, self.records["main_class"], "next")
                    except:
                        pass
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "synchronized_statement":
                self.edge_first_line(node_key, node_value)
                next_dest_index, next_node = self.get_next_index(node_value)
                last_line, line_type = self.get_block_last_line(node_value, "body")
                last_line_index = self.get_index(last_line)
                # if line_type in self.statement_types["non_control_statement"]:
                #     self.add_edge(last_line_index, 2, "exit_next")
                # Add an edge from start of synchronized to the line after it to signify async execution
                self.handle_next(node_value, next_node, "sync_next")
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "labeled_statement":
                self.edge_first_line(node_key, node_value)
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "if_statement":
                # Find the if block body and the else block body if exists (first statement inside them, add an edge)
                # Find the line just after the entire if_statement
                next_dest_index, next_node = self.get_next_index(node_value)
                # consequence
                self.edge_to_body(node_key, node_value, "consequence", "pos_next")
                # Find the last line in the consequence block and add an edge to the next statement
                last_line, line_type = self.get_block_last_line(node_value, "consequence")
                last_line_index = self.get_index(last_line)
                # Also add an edge from the last guy to the next statement after the if
                if line_type in self.statement_types["non_control_statement"]:
                    self.handle_next(last_line, next_node, "next_line 2")
                
                empty_if = False
                if last_line_index == current_index:
                    empty_if = True
                    self.handle_next(node_value, next_node, "next_line 3")
                if node_value.child_by_field_name("alternative") is not None:
                    # alternative
                    self.edge_to_body(node_key, node_value, "alternative", "neg_next")
                    # Find the last line in the alternative block
                    last_line, line_type = self.get_block_last_line(node_value, "alternative")
                    if line_type in self.statement_types["non_control_statement"]:
                        self.handle_next(last_line, next_node, "next_line 4")

                    if empty_if and last_line_index == current_index:
                        self.handle_next(node_value, next_node, "next_line 5")
                else:
                    # When else is not there add a direct edge from if node to the next statement
                    # if next_node is not None:
                    self.handle_next(node_value, next_node, "next_line 6")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type in self.statement_types["loop_control_statement"]:
                # Get the node immediately after the while statement
                next_dest_index, next_node = self.get_next_index(node_value)

                # Add an edge from this node to the first line in the body
                self.edge_to_body(node_key, node_value, "body", "pos_next")

                # Find the last line in the body block
                last_line, line_type = self.get_block_last_line(node_value, "body")
                last_line_index = self.get_index(last_line)
                # Add an edge from this node to the next line after the loop statement
                self.handle_next(node_value, next_node, "neg_next")
                # Add an edge from the last statement in the body to this node
                if line_type in self.statement_types["non_control_statement"]:
                    self.add_edge(last_line_index, current_index, "loop_control")

                # Add a self loop in case of for loops
                if current_node_type != "while_statement":
                    self.add_edge(current_index, current_index, "loop_update")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "do_statement":
                # Find the corresponding while statement
                # Add an edge from this node to the first line in the body
                # Find the last statement in the body and an edge frpom last line to the while node
                # Add an edge from the while node to the first line in the block or to the current do node
                # Add an edge from the while node to the next statement after the do_statement

                # Get the node immediately after the while statement
                next_dest_index, next_node = self.get_next_index(node_value)

                # Add an edge from this node to the first line in the body
                # if next_node is not None:
                self.edge_to_body(node_key, node_value, "body", "pos_next")

                # Find the last line in the body block
                last_line, line_type = self.get_block_last_line(node_value, "body")
                last_line_index = self.get_index(last_line)
                # Search the CFG_node_list for parameterized_expression with parent do_statement with AST_id = src_node
                while_index = 0
                while_node = None
                
                for k, v in node_list.items():
                    if (k[2] == "parenthesized_expression" and self.get_index(v.parent) == current_index):
                        while_index = self.index[k]
                        while_node = v
                        # break
                # Find the last statement in the body and an edge from last line to the while node
                self.add_edge(last_line_index, while_index, "next")

                # Add an edge from the while node to the first line in the block or to the current do node
                # self.CFG_edge_list.append((while_node, dest_node, 'loop_control')) # First node of block
                self.add_edge(while_index, current_index, "loop_control")  # do node

                # Add an edge from the while node to the next statement after the do_statement
                self.handle_next(while_node, next_node, "neg_next")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "continue_statement":
                # Go up the parent chain until we reach a loop_control statement or do statement
                # add edge from this node to the loop_control statement or do statement

                parent_node = node_value.parent
                loop_node = None
                while parent_node is not None:
                    if (
                        parent_node.type
                        in self.statement_types["loop_control_statement"]
                        or parent_node.type == "do_statement"
                    ):
                        loop_node = parent_node
                        break
                    parent_node = parent_node.parent

                # add edge from this node to the loop_control statement or do statement

                src_node = self.index[node_key]
                dest_node = self.get_index(loop_node)
                label_name = list(filter(lambda child: child.type == "identifier", node_value.children))
                if len(label_name) > 0:
                    target_name = label_name[0].text.decode("UTF-8") + ":"
                    dest_node = self.index[self.records["label_statement_map"][target_name]]  # Check on this later
                self.CFG_edge_list.append((src_node, dest_node, "jump_next"))

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "break_statement":
                # break with a label
                label_name = list(filter(lambda child: child.type == "identifier", node_value.children))
                if len(label_name) > 0:
                    target_name = label_name[0].text.decode("UTF-8") + ":"
                    # next_dest_index = self.records['label_statement_map'][target_name]
                    label_key = self.records["label_statement_map"][target_name]
                    label_node = node_list[label_key]
                    next_dest_index, next_node = self.get_next_index(label_node)
                    # need to get the node corresponding to this index, and then get the next node
                    self.handle_next(node_value, next_node, "jump_next")
                else:
                    # if it is inside a switch, it is handled here and also refer to switch_expression,
                    # if it is inside a loop, handle it here
                    parent_node = node_value.parent
                    loop_node = None
                    while parent_node is not None:
                        if (
                            parent_node.type
                            in self.statement_types["loop_control_statement"]
                            or parent_node.type == "do_statement"
                            or parent_node.type == "switch_expression"
                        ):
                            loop_node = parent_node
                            break
                        parent_node = parent_node.parent

                    next_dest_index, next_node = self.get_next_index(loop_node)
                    self.handle_next(node_value, next_node, "jump_next")
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "yield_statement":
                # if it is inside a switch, it is handled here and also refer to switch_expression,
                parent_node = node_value.parent
                loop_node = None
                while parent_node is not None:
                    if (
                        parent_node.type
                        in self.statement_types["loop_control_statement"]
                        or parent_node.type == "do_statement"
                        or parent_node.type == "switch_expression"
                    ):
                        loop_node = parent_node
                        break
                    parent_node = parent_node.parent

                try:
                    next_node = loop_node
                    next_dest_index = self.get_index(loop_node)
                except:
                    # Handle yield when no loop parent or switch parent
                    next_dest_index, next_node = self.get_next_index(node_value)
                self.handle_next(node_value, next_node, "yield_exit")
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "lambda_expression":
                self.edge_to_body(node_key, node_value, "body", "lambda_next")

            elif current_node_type == "switch_expression":
                # First check if the switch expression_statement is part of a non-control statement, and add an edge to the next line
                switch_parent = java_nodes.return_switch_parent_statement(node_value, self.statement_types["non_control_statement"])
                if switch_parent is not None:
                    try:
                        next_node = switch_parent.next_named_sibling
                        src_node = self.index[node_key]
                        dest_node = self.get_index(next_node)
                        if dest_node in self.records["switch_child_map"].keys():
                            dest_node = self.records["switch_child_map"][dest_node]
                        self.add_edge(src_node, dest_node, "next_line 7")

                    except Exception as e:
                        pass

                # Find all the case blocks associated with this switch node and add an edge to each of them
                # Find the last line in each case block and add an edge to the next case block unless it is a break statement
                # BUt if the block is empty, add an edge to the next case label
                # in case of a break statement, add an edge to the next statement outside the switch
                # in case of default, add an edge to the next statement outside the switch
                # in case of no default, add an edge from last block to the next statement outside the switch
                case_node_list = {}
                default_exists = False
                # default_list = []

                # Find the next statement outside the switch
                next_dest_index, next_node = self.get_next_index(node_value)

                # For each case label block, find the first statement in the block and add an edge to it
                for k, v in node_list.items():
                    if (k[2] == "switch_block_statement_group" or k[2] == "switch_rule") and self.get_index(v.parent.parent) == current_index:
                        case_node_list[k] = v
                        case_node_index = self.index[k]
                        self.add_edge(current_index, case_node_index, "switch_case")
                        case_label = list(filter(lambda child: child.type == "switch_label", v.children))
                        for case in case_label:
                            default_case = list(filter(lambda child: child.type == "default", case.children))
                            if len(default_case) > 0:
                                default_exists = True
                # If no default exists, add an edge from switch node to the next line after it
                if not default_exists:
                    self.handle_next(node_value, next_node, "switch_exit")

                for k, v in case_node_list.items():
                    current_case_index = self.index[k]
                    current_case = v
                    case_statements = list(filter(lambda child: (child.is_named and child.type != "switch_label"),v.children))
                    case_labels = list(filter(lambda child: (child.is_named and child.type == "switch_label"),v.children))
                    next_case_node = v.next_named_sibling
                    try:
                        next_case_node_index = self.get_index(next_case_node)
                    except:
                        next_case_node_index = None

                    block_node = None
                    if len(case_statements) == 0:
                        default_flag = False
                        # Check if its a default statement, then add an edge to the next statement outside the switch
                        for case_label in case_labels:
                            default_case = list(
                                filter(
                                    lambda child: child.type == "default",
                                    case_label.children,
                                )
                            )
                            if len(default_case) > 0:
                                self.handle_next(current_case, next_node, "default_exit")
                                default_flag = True
                                break
                        # There is no case body, so add an edge from this case to the next case label, if exists
                        if not default_flag and next_case_node_index is not None:
                            self.add_edge(current_case_index, next_case_node_index, "fall through")

                    else:
                        # The case body exists
                        empty_block = False
                        if (
                            len(case_statements) > 1
                            and case_statements[0].type == "block"
                            and len(list(filter(lambda child: child.is_named, case_statements[0].children))) == 0
                        ):
                            empty_block = True
                            # Empty block followed by other statement in the same case
                            next_line = case_statements[1]
                            next_line_index = self.get_index(next_line)
                            self.add_edge(current_case_index, next_line_index, "next_line 8")

                        # Find the first line in each case block and add an edge from case label to itandl empty case block
                        
                        explicit_block = None
                        for child in v.children:
                            # Need to write a loop for unlimited layers of nesting
                            if child.is_named and child.type != "switch_label":
                                if child.type == "block":
                                    explicit_block = child
                                    for child in child.children:
                                        if (
                                            child.is_named
                                            and child.type != "switch_label"
                                        ):
                                            block_node = child
                                            break
                                else:
                                    block_node = child
                                break
                        if block_node is None:
                            # Add fall through edge to next case label
                            if empty_block == False and next_case_node_index is not None:
                                self.add_edge(current_case_index, next_case_node_index, "fall through")
                        else:
                            first_line_index = self.get_index(block_node)
                            self.add_edge(current_case_index, first_line_index, "case_next")

                        # if the case contains a set of braces and the break outside it,
                        # then add an edge from the last line inside the block to the first line outside it
                        if explicit_block is not None:
                            next_line = explicit_block.next_named_sibling
                            if next_line is not None:
                                next_line_index = self.get_index(next_line)
                                block_node = None
                                for child in reversed(explicit_block.children):
                                    # Need to write a loop for unlimited layers of nesting
                                    if child.is_named:
                                        if child.type == "block":
                                            for child in reversed(child.children):
                                                if child.is_named:
                                                    block_node = child
                                                    break
                                        else:
                                            block_node = child
                                        break
                                if block_node is not None:
                                    last_line_index = self.get_index(block_node)
                                    if (block_node.type in self.statement_types["non_control_statement"]):
                                        self.add_edge(last_line_index, next_line_index, "fall through")

                        # Find the last line in each case block and add an edge to the next case block unless it is a break statement
                        block_node = None
                        for child in reversed(v.children):
                            # Need to write a loop for unlimited layers of nesting
                            if child.is_named:
                                if child.type == "block":
                                    for child in reversed(child.children):
                                        if child.is_named:
                                            block_node = child
                                            break
                                else:
                                    block_node = child
                                break
                        if block_node is not None:
                            last_line_index = self.get_index(block_node)
                            if block_node.type in self.statement_types["non_control_statement"] and next_case_node_index is not None:
                                self.add_edge(
                                    last_line_index,
                                    next_case_node_index,
                                    "fall through",
                                )

                        # in case of default, add an edge to the next statement outside the switch
                        # in case of no default, add an edge from last block to the next statement outside the switch
                    if next_case_node_index is None:
                        # This is the last block
                        if block_node is not None:
                            if (block_node.type in self.statement_types["non_control_statement"]):
                                self.handle_next(block_node, next_node, "switch_out")
                        else:
                            self.handle_next(v, next_node, "switch_out")

                # in case of a break statement, add an edge to the next statement outside the switch
                # -> Handled in break statement
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "return_statement":
                # Check if it returns a function with an inner definition
                # method_list will give us the index of the inner function
                inner_definition = []
                self.returns_inner_definition(node_value, inner_definition)
                
                if len(inner_definition) > 0:
                    inner_definition = inner_definition[0] 
                    # inner_index = self.index[(inner_definition.start_point, inner_definition.end_point, inner_definition.type)]
                    # TODOD: Perhaps remove this because inconsistent wih newly decided logic
                    self.add_edge(current_index, inner_definition, "return_next")

                if not self.inner_function(node_value):
                    # pass
                    # Instead of an edge to exit node, we update the return map
                    # self.add_edge(current_index, 2, "return_exit")
                    containing_method = self.get_index(self.get_containing_method(node_value))
                    
                    try:
                        self.records["return_statement_map"][containing_method].append(current_index)
                    except:
                        self.records["return_statement_map"][containing_method] = [current_index]
                # else: TODO: When inner functions are properly implemented
                #       save it in the return map?
                #     next_dest_index, next_node = self.get_next_index(node_value)
                #     self.add_edge(current_index, next_dest_index, "return_exit")
            # ------------------------------------------------------------------------------------------------
            # elif current_node_type == "assert_statement":
            #     self.add_edge(current_index, 2, "assert_false")
            # ------------------------------------------------------------------------------------------------
            elif (
                current_node_type == "try_statement"
                or current_node_type == "try_with_resources_statement"
            ):
                # Add edge from try block to first statement inside the block
                self.edge_to_body(node_key, node_value, "body", "next")
                catch_node_list = {}
                finally_node = None
                # Identify all catch blocks - add an edge from catch node to first line inside block
                for k, v in node_list.items():

                    if (k[2] == "catch_clause" and self.get_index(v.parent) == current_index):
                        catch_node_list[k] = v
                        self.edge_to_body(k, v, "body", "next")
                    elif (k[2] == "finally_clause" and self.get_index(v.parent) == current_index):
                        finally_node = v
                        self.edge_first_line(k, v)  # Not sure if this works

                # From each line inside the try block, an edge going to each catch block
                try_body = node_value.child_by_field_name("body")
                statements = []
                statements = self.get_all_statements(try_body, node_list, statements)
                if len(statements) > 0:
                    for statement in statements:
                        statement_index = self.get_index(statement)
                        for catch_node in catch_node_list.keys():
                            catch_index = self.index[catch_node]
                            # if statement.type != 'return_statement': The Exception can occur on the RHS so the catch_exception edge should be there
                            self.add_edge(statement_index, catch_index, "catch_exception")

                # Find the next statement outside the try block
                next_dest_index, next_node = self.get_next_index(node_value)
                exit_next = None

                # finally block is optional
                if finally_node is not None:
                    # From last line of finally to next statement outside the try block
                    last_line, line_type = self.get_block_last_line(finally_node, "body")
                    if line_type in self.statement_types["non_control_statement"]:
                        self.handle_next(last_line, next_node, "finally_exit")
                    # For the remaining portion, set finally block as next node if exists
                    exit_next = finally_node
                else:
                    exit_next = next_node

                # From last line of try block to finally or to next statement outside the try block
                last_line, line_type = self.get_block_last_line(node_value, "body")
                # if line_type in self.statement_types['non_control_statement']: # Inside try, edges to finally should be irrespective of control or non control statements
                if line_type in self.statement_types["node_list_type"]:
                    self.handle_next(last_line, exit_next, "try_exit")
                # From last line of every catch block to finally or to next statement outside the try block
                for catch_node, catch_value in catch_node_list.items():
                    last_line, line_type = self.get_block_last_line(catch_value, "body")
                    last_line_index = self.get_index(last_line)
                    if line_type in self.statement_types["non_control_statement"]:
                        self.handle_next(last_line, exit_next, "catch_exit")
                        # Case of empty catch block
                    elif last_line_index == self.index[catch_node]:
                        self.handle_next(last_line, exit_next, "catch_exit")
                        # ------------------------------------------------------------------------------------------------
            elif current_node_type == "throw_statement":
                # Control goes to the first catch in the call stack (dynamic call stack) Nothing to do statically
                # Essentially exits the function
                parent = node_value.parent
                # try_flag = False
                try_node = None
                while parent is not None:
                    if parent.type == "catch_clause" or parent.type == "finally_clause":
                        break
                    if (
                        parent.type == "try_statement"
                        or parent.type == "try_with_resources_statement"
                    ):
                        try_flag = True
                        try_node = parent
                        break
                    parent = parent.parent
                if try_node is None:
                    # self.add_edge(current_index, 2, "throw_exit")
                    pass
                else:
                    # try_node is like the current_node, helps us find the corresponding catch clauses

                    # Add edge from try block to first statement inside the block
                    # self.edge_to_body(node_key, node_value, 'body', 'next')
                    catch_node_list = {}
                    # finally_node = None
                    # Identify all catch blocks
                    for k, v in node_list.items():
                        if ( k[2] == "catch_clause" and self.get_index(v.parent) == current_index):
                            catch_node_list[k] = v
                    # if len(catch_node_list) == 0:
                    #     self.add_edge(current_index, 2, "throw_exit")
        
        self.add_method_call_edges()
        if warning_counter > 0:
            print("Total number of warnings from unimplemented statement types: ", warning_counter)
        return self.CFG_node_list, self.CFG_edge_list



from ..tree_parser.custom_parser import CustomParser

class JavaParser(CustomParser):
    def __init__(self, src_language, src_code):
        super().__init__(src_language, src_code)

    def check_declaration(self, current_node):

        # node_list_type = ['local_variable_declaration','declaration', 'expression_statement', 'labeled_statement', 'if_statement', 'while_statement', 'for_statement', 'enhanced_for_statement', 'assert_statement', 'do_statement', 'break_statement', 'continue_statement', 'return_statement', 'yield_statement', 'switch_expression', 'synchronized_statement', 'local_variable_declaration', 'throw_statement', 'try_statement', 'try_with_resources_statement', 'method_declaration','constructor_declaration', 'switch_block_statement_group', 'switch_rule', 'throw_statement', 'explicit_constructor_invocation']
        parent_types = ["variable_declarator", "catch_formal_parameter", "formal_parameter"]
        current_types = ["identifier"]
        if (
            current_node.parent is not None
            and current_node.parent.type in parent_types
            and current_node.type in current_types
        ):
            if current_node.parent.type == "variable_declarator":
                # TODO: If it is a class attribute need to maintain class_name.class_variable so that the correct datatype can be mapped
                if len(list(filter(lambda x: x.type == "=", current_node.parent.children)))> 0:
                    if current_node.next_sibling is not None and current_node.next_sibling.type == "=":
                        return True
                    else:
                        return False
                else:
                    return True
            return True
        return False
    
    def get_type(self, node):
        """ Given a variable declarator node, return the variable type of the identifier"""
        datatypes = ['type_identifier', 'integral_type', 'floating_point_type', 'void_type', 'boolean_type', 'scoped_type_identifier', 'generic_type']
        if node.type == "formal_parameter":
            return node.children[0].text.decode('utf-8')
        
        for child in node.parent.children:
            if child.type in datatypes:
                return child.text.decode('utf-8')
        return None
    
    def scope_check(self, parent_scope, child_scope):
        for p in parent_scope:
            if p not in child_scope:
                return False
        return True
    def longest_scope_match(self, name_matches, symbol_table):
        """Given a list of name matches, return the longest scope match"""
        # [(ind, var), (ind,var)]
        scope_array = list(map(lambda x: symbol_table['scope_map'][x[0]], name_matches))
        # scope_array.sort(key=lambda x: len(x), reverse=True)
        # index = max(range(len(scope_array)), key=lambda x: len(x))
        max_val = max(scope_array, key=lambda x: len(x))
        for i in range(len(scope_array)):
            if scope_array[i] == max_val:
                return name_matches[i][0]
        


    def create_all_tokens(
        self,
        src_code,
        root_node,
        all_tokens,
        label,
        method_map,
        method_calls,
        start_line,
        declaration,
        declaration_map,
        symbol_table,
    ):
        # Needs to be modifed for every language
        remove_list = ["class_declaration", "method_declaration", "method_invocation"]
        # node_list_type = ['local_variable_declaration', 'declaration', 'expression_statement', 'labeled_statement',
        #                   'if_statement', 'while_statement', 'for_statement', 'enhanced_for_statement',
        #                   'assert_statement', 'do_statement', 'break_statement', 'continue_statement',
        #                   'return_statement', 'yield_statement', 'synchronized_statement',
        #                   'local_variable_declaration', 'throw_statement', 'try_statement',
        #                   'try_with_resources_statement', 'method_declaration', 'constructor_declaration',
        #                   'switch_expression', 'switch_block_statement_group', 'switch_rule', 'throw_statement',
        #                   'explicit_constructor_invocation']
        block_types = [
            "block",
            "if_statement",
            "while_statement",
            "for_statement",
            "enhanced_for_statement",
            "do_statement",
            "switch_expression",
            "switch_block_statement_group",
            "switch_rule",
            "throw_statement",
            "synchronized_statement",
            "try_statement",
            "try_with_resources_statement",
            "method_declaration",
            "constructor_declaration",
            "explicit_constructor_invocation",
        ]

        if root_node.is_named and root_node.type in block_types:
            """On entering a new block, increment the scope id and push it to the scope_stack"""
            # current_scope = symbol_table['scope_stack'][-1]
            symbol_table["scope_id"] = symbol_table["scope_id"] + 1
            symbol_table["scope_stack"].append(symbol_table["scope_id"])

        if (
            root_node.is_named
            and (len(root_node.children) == 0 or root_node.type == "string")
            and root_node.type != "comment"
        ): # All identifiers
            index = self.index[(root_node.start_point, root_node.end_point, root_node.type)]
            label[index] = root_node.text.decode("UTF-8")
            start_line[index] = root_node.start_point[0]
            all_tokens.append(index)
            """Store a copy of the current scope stack in the scope map for each token. This token belongs to all the scopes in the current scope stack."""
            
            symbol_table["scope_map"][index] = symbol_table["scope_stack"].copy()

            
            current_node = root_node
            if (current_node.parent is not None and current_node.parent.type in remove_list):
                method_map.append(index)
                if current_node.next_named_sibling.type == "argument_list":
                    method_calls.append(index)

            if (current_node.parent is not None and current_node.parent.type == "field_access"):
                object_node = current_node.parent.child_by_field_name("field")
                object_index = self.index[(object_node.start_point, object_node.end_point, object_node.type)]
                current_index = self.index[(current_node.start_point,current_node.end_point,current_node.type)]
                if object_index == current_index:
                    method_map.append(current_index)

                while (current_node.parent is not None and current_node.parent.type == "field_access"):
                    current_node = current_node.parent

                if (current_node.parent is not None and current_node.parent.type == "method_invocation"):
                    method_map.append(index)
                label[index] = current_node.text.decode("UTF-8")
                
            if self.check_declaration(current_node):
                variable_name = label[index]
                declaration[index] = variable_name

                variable_type = self.get_type(current_node.parent)
                if variable_type is not None:
                    symbol_table["data_type"][index] = variable_type
            else:
                current_scope = symbol_table['scope_map'][index]

                if current_node.type == "field_access":
                    
                    field_variable = current_node.children[-1]
                    # entire_variable_name = current_node.text.decode('utf-8')
                    field_variable_name = field_variable.text.decode('utf-8')

                    for (ind,var) in declaration.items():
                        if var == field_variable_name:
                            parent_scope = symbol_table['scope_map'][ind]
                            if self.scope_check(parent_scope, current_scope):
                                declaration_map[index] = ind
                                break

                    
                        # Identify the leftmost - that is the reference - identify its class
                # TODO: Handle field accesses
                # If we can add the coorect declaration_map entry, we should be done.
                # For this we need to identify the class to which it belongs, and the corresponding field variable
                else:
                    name_matches = []
                    for (ind, var) in declaration.items():
                        if var == label[index]:
                            parent_scope = symbol_table['scope_map'][ind]
                            if self.scope_check(parent_scope, current_scope):
                                name_matches.append((ind,var))
                    for (ind, var) in name_matches:
                        parent_scope = symbol_table['scope_map'][ind]
                        closest_index = self.longest_scope_match(name_matches, symbol_table)
                        declaration_map[index] = closest_index
                        break

        else:
            for child in root_node.children:
                self.create_all_tokens(
                    src_code,
                    child,
                    all_tokens,
                    label,
                    method_map,
                    method_calls,
                    start_line,
                    declaration,
                    declaration_map,
                    symbol_table,
                )

        if root_node.is_named and root_node.type in block_types:
            """On exiting a block, pop the scope id from the scope_stack"""
            symbol_table["scope_stack"].pop(-1)

        return (
            all_tokens,
            label,
            method_map,
            method_calls,
            start_line,
            declaration,
            declaration_map,
            symbol_table,
        )



# from codeviews.CFG.CFG_python import CFGGraph_python
from .CFG_csharp import CFGGraph_csharp
from .CFG_java import CFGGraph_java
from ...tree_parser.parser_driver import ParserDriver
from ...utils import postprocessor


class CFGDriver:
    def __init__(
        self,
        src_language="java",
        src_code="",
        output_file="CFG_output.json",
        properties={},
    ):
        self.src_language = src_language

        self.parser = ParserDriver(src_language, src_code).parser
        self.root_node = self.parser.root_node
        self.src_code = self.parser.src_code
        self.properties = properties

        self.CFG_map = {
            "java": CFGGraph_java,
            "cs": CFGGraph_csharp,
            # "python": CFGGraph_python
        }

        self.CFG = self.CFG_map[self.src_language](
            self.src_language,
            self.src_code,
            self.properties,
            self.root_node,
            self.parser,
        )
        self.node_list = self.CFG.node_list
        self.graph = self.CFG.graph
        if output_file:
            self.json = postprocessor.write_networkx_to_json(self.graph, output_file)
            postprocessor.write_to_dot(
                self.graph, output_file.split(".")[0] + ".dot", output_png=True
            )



from ..tree_parser.custom_parser import CustomParser


class CSParser(CustomParser):
    def __init__(self, src_language, src_code):
        super().__init__(src_language, src_code)

    def check_declaration(self, current_node):
        parent_types = ["variable_declarator", "catch_declaration"]
        current_types = ["identifier"]
        if (
                current_node.parent is not None
                and current_node.parent.type in parent_types
                and current_node.type in current_types
        ):
            return True
        return False

    def get_type(self, node):
        """ Given a variable declarator node, return the variable type of the identifier"""
        datatypes = ["alias_qualified_name", "array_type", "function_pointer_type",
                     "generic_name", "global", "identifier", "implicit_type", "nullable_type",
                     "pointer_type", "predefined_type", "qualified_name", "ref_type", "scoped_type",
                     "tuple_type"]
        # TODO: Triage this
        # print(node.type, node.text.decode('utf-8'))
        if node.type == "parameter":
            return node.children[0].text.decode('utf-8')

        for child in node.parent.children:
            # print(child.type, child.text.decode('utf-8'))
            if child.type in datatypes:
                return child.text.decode('utf-8')
        return None

    def scope_check(self, parent_scope, child_scope):
        for p in parent_scope:
            if p not in child_scope:
                # print("parent", parent_scope, "child", child_scope)
                return False
        return True

    def create_all_tokens(
            self,
            src_code,
            root_node,
            all_tokens,
            label,
            method_map,
            method_calls,
            start_line,
            declaration,
            declaration_map,
            symbol_table,
    ):
        # Needs to be modifed for every language

        remove_list = [
            "class_declaration",
            "method_declaration",
            "invocation_expression",
            # ? Not sure if these should go in this logic
            # 'constructor_initializer', 'implicit_object_creation_expression', 'object_creation_expression', 'primary_constructor_base_type'
        ]
        block_types = [
            "block",
            "checked_statement",
            "fixed_statement",
            "unsafe_statement",
            "using_statement",
            "if_statement",
            "while_statement",
            "for_statement",
            "for_each_statement",
            "do_statement",
            "switch_expression",
            "switch_expression_arm",
            "switch_section",
            "switch_statement",
            "try_statement",
            "method_declaration",
            "constructor_declaration",
            "property_declaration",
            "interface_declaration",
            "throw_statement",
        ]
        # For reach identifier, if the parent is field_access -> keep going up till you reach the highest parent
        # else map to self
        # check if parent of highest mapped identifier is method invocation, then add it to the list
        # This logic helps create more meaningful tokens when it comes to variable names of object oriented languages like Java

        # if root_node.is_named and root_node.type

        # if root_node.type == '{':
        #     print(root_node.type)
        #     # current_scope = symbol_table['scope_stack'][-1]
        #     symbol_table['scope_id'] = symbol_table['scope_id'] + 1
        #     symbol_table['scope_stack'].append(symbol_table['scope_id'])

        # elif root_node.type == '}':
        #     print(root_node.type)
        #     symbol_table['scope_stack'].pop(-1)

        if root_node.is_named and root_node.type in block_types:
            # print(root_node.type)
            # current_scope = symbol_table['scope_stack'][-1]
            symbol_table["scope_id"] = symbol_table["scope_id"] + 1
            symbol_table["scope_stack"].append(symbol_table["scope_id"])

        if (
                root_node.is_named
                and (len(root_node.children) == 0 or root_node.type == "string")
                and root_node.type != "comment"
        ):
            index = self.index[
                (root_node.start_point, root_node.end_point, root_node.type)
            ]
            label[index] = root_node.text.decode("UTF-8")
            # if label[index] == 'true':
            #     print("FOUND TRUEE")
            start_line[index] = root_node.start_point[0]
            all_tokens.append(index)
            symbol_table["scope_map"][index] = symbol_table["scope_stack"].copy()
            # print("Adding to scope map", index, symbol_table['scope_map'][index])

            current_node = root_node

            if (
                    current_node.parent is not None
                    and current_node.parent.type in remove_list
            ):
                method_map.append(index)
                # print(current_node.text.decode('utf-8'), current_node.next_named_sibling)
                if current_node.next_named_sibling.type == "argument_list":
                    # print("ADDING METHOD", current_node.type, current_node.text.decode('utf-8'))
                    method_calls.append(index)

            if (
                    current_node.parent is not None
                    and current_node.parent.type == "member_access_expression"
            ):
                object_node = current_node.parent.child_by_field_name("name")
                object_index = self.index[
                    (object_node.start_point, object_node.end_point, object_node.type)
                ]
                current_index = self.index[
                    (
                        current_node.start_point,
                        current_node.end_point,
                        current_node.type,
                    )
                ]
                if object_index == current_index:
                    # ? not sure about this yet Console.writeline - if i say expression it will pick console if i say name it will pick the object but java side it picks the parent?
                    method_map.append(current_index)

                while (
                        current_node.parent is not None
                        and current_node.parent.type == "member_access_expression"
                ):
                    current_node = current_node.parent

                if (
                        current_node.parent is not None
                        and current_node.parent.type == "invocation_expression"
                ):
                    method_map.append(index)
                label[index] = current_node.text.decode("UTF-8")

            # Modify this to take care of scope
            if self.check_declaration(current_node):
                variable_name = label[index]
                declaration[index] = variable_name

                variable_type = self.get_type(current_node.parent)
                if variable_type is not None:
                    symbol_table["data_type"][index] = variable_type
            else:
                current_scope = symbol_table['scope_map'][index]
                for (ind, var) in declaration.items():
                    if var == label[index]:
                        parent_scope = symbol_table['scope_map'][ind]
                        # print(ind,var)
                        if self.scope_check(parent_scope, current_scope):
                            # print(ind, var)
                            # print("^^", current_node.parent.type, index, var, "Current:", current_scope, "Parent:", parent_scope)
                            declaration_map[index] = ind
                            break

        else:
            for child in root_node.children:
                self.create_all_tokens(
                    src_code,
                    child,
                    all_tokens,
                    label,
                    method_map,
                    method_calls,
                    start_line,
                    declaration,
                    declaration_map,
                    symbol_table,
                )

        if root_node.is_named and root_node.type in block_types:
            symbol_table["scope_stack"].pop(-1)

        return (
            all_tokens,
            label,
            method_map,
            method_calls,
            start_line,
            declaration,
            declaration_map,
            symbol_table,
        )



import copy
import pprint
import time
from collections import defaultdict

import networkx as nx
from deepdiff import DeepDiff
from loguru import logger

from ...utils.java_nodes import statement_types
from ...utils.src_parser import traverse_tree

assignment = ["assignment_expression"]
def_statement = ["variable_declarator"]
increment_statement = ["update_expression"]
variable_type = ['identifier', 'this']
method_calls = ["method_invocation"]

switch_type = ['switch_expression']
inner_types = [
    "declaration",
    "expression_statement",
    "local_variable_declaration",
]
method_invocation = method_calls + ["object_creation_expression", "explicit_constructor_invocation"]
method_declaration = ['method_declaration', 'constructor_declaration']
handled_types = assignment + def_statement + increment_statement + method_calls + method_declaration + switch_type + [
    "catch_clause", "return_statement", "throw_statement", "enhanced_for_statement", "for_statement",
    "object_creation_expression", "lambda_expression"]

pp = pprint.PrettyPrinter(indent=4)
system_type = ("Console", "System", "String", "Collections")
# call_variable_map = {}
# node_list = {}
debug = False
# if any(
#         # GITHUB_ACTIONS
#         x in os.environ for x in ("PYCHARM_HOSTED",)
# ):
#     debug = any(
#         # GITHUB_ACTIONS
#         x in os.environ for x in ("PYCHARM_HOSTED",)
# ):
#     debug = True
# properties = {
#     "last_use": False,
#     "last_def": False,
#     "alex_algo": False
# }


def scope_check(parent_scope, child_scope):
    for p in parent_scope:
        if p not in child_scope:
            return False
    return True


class Identifier:
    def __init__(self, parser, node, line=None, declaration=False, full_ref=None, method_call=False):
        self.core = st(node)
        # field_declaration = return_first_parent_of_types(node, "field_declaration")
        # param_declaration = return_first_parent_of_types(node, "formal_parameter")
        # self.param_declaration = False
        # if param_declaration is not None:
        #     self.param_declaration = True
        # self.immutable_declaration = False
        # if field_declaration is not None:
        #     self.immutable_declaration = True
        class_node = return_first_parent_of_types(node, "class_declaration")
        if full_ref is None:
            full_ref = node
        self.parent_class = class_node
        self.unresolved_name = st(full_ref)
        if self.unresolved_name.startswith("this."):
            self.unresolved_name = st(full_ref)[5:]
        if class_node is not None:
            class_name = st(class_node.child_by_field_name("name"))
            self.parent_class = class_name
            if self.unresolved_name.startswith(class_name + "."):
                self.unresolved_name = self.unresolved_name[len(class_name) + 1:]
        self.line = line
        self.declaration = declaration
        self.method_call = method_call
        self.satisfied = method_call
        if self.method_call and self.declaration:
            method_resolution = self.unresolved_name.split(".")
            # br.attribute2.attr.method1
            # Option 1 - method call affects main object
            # resolved_name = br
            # resolved_name = method_resolution[0]
            # resolved_name = br.attribute2.attr
            # Option 2 - method call only affects subattribute
            resolved_name = ".".join(method_resolution[:-1])
            self.name = resolved_name
        else:
            self.name = self.unresolved_name
        if not self.name:
            self.name = self.unresolved_name
        if node.type == "field_access":
            node = recursively_get_children_of_types(node, ['identifier'], index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])[-1]
        variable_index = get_index(node, parser.index)
        self.variable_scope = parser.symbol_table["scope_map"][variable_index]
        if variable_index in parser.declaration_map:
            self.scope = parser.symbol_table["scope_map"][parser.declaration_map[variable_index]]
        else:
            # self.scope = variable_scope
            # Declaration does not exist hence it is given
            # the outermost scope
            self.scope = [0]
        if line is not None:
            self.real_line_no = read_index(parser.index, line)[0][0]

    def __eq__(self, other):
        # if not isinstance(other, Identifier):
        #     return NotImplemented
        # if self.name == other.name:
        #     if other.line is None and other.line is not None:
        #         other.line = self.line
        #     elif other.line is not None and self.line is None:
        #         self.line = other.line

        # if self.method_call and self.declaration:
        #     method_resolution = self.name.split(".")
        #     # Option 1 - method call affects main object
        #     # resolved_name = method_resolution[0]
        #     # Option 2 - method call only affects subattribute
        #     resolved_name = ".".join(method_resolution[:-1])
        #     return resolved_name == other.name and self.line == other.line and sorted(self.scope) == sorted(
        #         other.scope)
        # else:
        return self.name == other.name and self.line == other.line and sorted(self.scope) == sorted(
            other.scope) and self.method_call == other.method_call

    def __hash__(self):
        # if self.method_call and self.declaration:
        #     method_resolution = self.name.split(".")
        #     # Option 1 - method call affects main object
        #     resolved_name = method_resolution[0]
        #     # Option 2 - method call only affects subattribute
        #     # resolved_name = ".".join(method_resolution[:-1])
        #     return hash((resolved_name, self.line, str(self.scope), False))
        # else:
        return hash((self.name, self.line, str(self.scope), self.method_call))

    def __str__(self):
        result = [self.name]
        if self.line:
            result += [str(self.real_line_no)]
            result += ['|'.join(map(str, self.scope))]
        else:
            result += ["?"]
        if self.method_call:
            result += ["()"]
        return f"{{{','.join(result)}}}"


def st(child):
    if child is None:
        # logger.error
        return ""
    else:
        return child.text.decode()


def get_index(node, index):
    try:
        return index[(node.start_point, node.end_point, node.type)]
    except:
        return None


def read_index(index, idx):
    return list(index.keys())[list(index.values()).index(idx)]


def parent_remapping_callback(node, remap=True):
    if node is None:
        return None
    elif node.type == "do_statement":
        if not remap:
            return None
        while_node = None
        for child in node.children:
            if child.type == "while":
                while_node = child.next_named_sibling
                break
        return while_node
    elif node.type == "try_statement":
        return None
    else:
        return node


def return_first_parent_of_types(node, parent_types, stop_types=None, remap=True):
    if stop_types is None:
        stop_types = []
    if node.type in parent_types:
        return parent_remapping_callback(node, remap=remap)
    while node.parent is not None:
        if node.type in stop_types:
            return None
        if node.parent.type in parent_types:
            return parent_remapping_callback(node.parent, remap=remap)
        node = node.parent
    return None


#
# def set_add(lst, item):
#     for entry in lst:
#         if item == entry:
#             return
#     lst.append(item)
#
#
# def set_union(first_list, second_list):
#     resulting_list = list(first_list)
#     for x in second_list:
#         set_add(resulting_list, x)
#     return resulting_list
#
#
# def set_difference(first_list, second_list):
#     resulting_list = []
#     for item in first_list:
#         match_found = False
#         for item2 in second_list:
#             if item == item2:
#                 match_found = True
#                 break
#         if not match_found:
#             resulting_list.append(item)
#     return resulting_list
#
#
# def set_intersection(first_list, second_list):
#     resulting_list = []
#     for item in first_list:
#         match_found = False
#         for item2 in second_list:
#             if item == item2:
#                 match_found = True
#                 break
#         if match_found:
#             resulting_list.append(item)
#     return resulting_list


def set_add(lst, item):
    for entry in lst:
        if item == entry:
            return
    lst.append(item)


def set_union(first_list, second_list):
    resulting_list = list(first_list)
    resulting_list.extend(x for x in second_list if x not in resulting_list)
    return resulting_list


def set_difference(first_list, second_list):
    resulting_list = [item for item in first_list if item not in second_list]
    return resulting_list


def set_intersection(first_list, second_list):
    resulting_list = [item for item in first_list if item in second_list]
    return resulting_list


def add_entry(parser, rda_table, statement_id, used=None, defined=None, declaration=False, core=None,
              method_call=False):
    # global node_list
    if statement_id not in rda_table:
        rda_table[statement_id] = defaultdict(list)
    if not used and not defined:
        return
    mapped_node = None
    # meth_statement = return_first_parent_of_types(used or defined, ['method_declaration', 'constructor_declaration'])
    # if meth_statement is not None:
    #     meth_statement_id = get_index(meth_statement, parser.index)
    #     if meth_statement_id in call_variable_map and meth_statement_id != statement_id:
    #         checker = st(core).split(".")[0] or st(defined).split(".")[0] or st(used).split(".")[0]
    #         if checker in call_variable_map[meth_statement_id]:
    #             mapped_node = call_variable_map[meth_statement_id][checker]
    #             # if mapped_node.parent.type == "field_access":
    #             #     mapped_node = recursively_get_children_of_types(mapped_node, ['identifier'], index=parser.index,
    #             #                                                      check_list=parser.symbol_table["scope_map"])[-1]
    if core is None:
        current_node = used or defined
        if current_node.type == "field_access":
            current_node = recursively_get_children_of_types(current_node, ['identifier'], index=parser.index,
                                                             check_list=parser.symbol_table["scope_map"])[-1]
            if defined is not None:
                defined = current_node
            else:
                used = current_node
        if (
                (current_node.parent is not None
                 and current_node.parent.type == "field_access")
        ):
            # LOOK AT THE SYMBOL TABLE COMMENT AS WELL #? TODO
            # object_node = current_node.parent.child_by_field_name("name")
            # object_index = get_index(object_node, parser.index)
            # current_index = get_index(current_node, parser.index)
            # if object_index == current_index:
            #     all_calls.append(current_index)

            while (
                    current_node.parent is not None
                    and current_node.parent.type == "field_access"
            ):
                current_node = current_node.parent

            if (
                    current_node.parent is not None
                    and current_node.parent.type == "method_invocation"
            ):
                if not st(current_node).startswith(system_type):
                    add_entry(parser, rda_table, statement_id, defined=used or defined, core=current_node.parent,
                              # declaration=True,
                              method_call=True)
                    add_entry(parser, rda_table, statement_id, used=used or defined, core=current_node.parent,
                              # declaration=True,
                              method_call=True)
                if mapped_node is not None:
                    set_add(rda_table[statement_id]["def"],
                            Identifier(parser, mapped_node, statement_id, full_ref=mapped_node, declaration=declaration,
                                       method_call=method_call))
                return
            if defined is not None:
                add_entry(parser, rda_table, statement_id, defined=defined, core=current_node, declaration=declaration,
                          method_call=method_call)
                if mapped_node is not None:
                    set_add(rda_table[statement_id]["def"],
                            Identifier(parser, mapped_node, statement_id, full_ref=mapped_node, declaration=declaration,
                                       method_call=method_call))
            else:
                add_entry(parser, rda_table, statement_id, used=used, core=current_node, declaration=declaration,
                          method_call=method_call)
            return
        elif current_node.next_sibling and current_node.next_sibling.type == "." and current_node.parent is not None and current_node.parent.type == "method_invocation":
            if not st(current_node).startswith(system_type):
                if (used or defined).type == "identifier":
                    set_add(rda_table[statement_id]["def"],
                            Identifier(parser, used or defined, statement_id, full_ref=None,
                                       # declaration=True,
                                       method_call=True))
    if st(core).startswith(system_type):
        return
    if defined is not None:
        if get_index(defined, parser.index) not in parser.symbol_table["scope_map"]:
            return
        set_add(rda_table[statement_id]["def"],
                Identifier(parser, defined, statement_id, full_ref=core, declaration=declaration,
                           method_call=method_call))
        if mapped_node is not None:
            set_add(rda_table[statement_id]["def"],
                    Identifier(parser, mapped_node, statement_id, full_ref=mapped_node, declaration=declaration,
                               method_call=method_call))
    else:
        if get_index(used, parser.index) not in parser.symbol_table["scope_map"]:
            return
        if mapped_node is not None and method_call:
            set_add(rda_table[statement_id]["def"],
                    Identifier(parser, mapped_node, statement_id, full_ref=mapped_node, declaration=declaration,
                               method_call=method_call))
        set_add(rda_table[statement_id]["use"],
                Identifier(parser, used, full_ref=core, declaration=declaration, method_call=method_call))


def check_field_name(req_child):
    for i, child in enumerate(req_child.parent.children):
        if req_child == child:
            field_type = child.parent.field_name_for_child(i)
            if field_type is not None and field_type == "type":
                return False
    return True


def recursively_get_children_of_types(
        node, st_types, check_list=None, index=None, result=None, stop_types=None
):
    if isinstance(st_types, str):
        st_types = [st_types]
    if stop_types is None:
        stop_types = []
    if result is None:
        result = []
    if node.type in stop_types:
        return result
    if check_list and index:
        result.extend(
            list(
                filter(
                    lambda child: child.type in st_types
                                  and get_index(child, index) in check_list
                                  and check_field_name(child),
                    # and child.parent.type in ['variable_declaration', 'object_creation_expression',
                    #                           'field_access'],
                    node.children,
                )
            )
        )
    else:
        result.extend(list(filter(lambda child: child.type in st_types, node.children)))
    for child in node.named_children:
        if child in stop_types:
            continue
        # if child.type not in st_types:
        result = recursively_get_children_of_types(
            child, st_types, result=result, stop_types=stop_types, index=index, check_list=check_list
        )
    return result


def compute_out(old_result, connected_derivers, def_set):
    result = []
    for node in connected_derivers:
        result = set_union(result, old_result[node]["IN"])
    # for node in result:
    #     if node in def_set:
    #         continue
    return result


def compute_in(use_set, out_set, def_set, all_defs):
    result = set_union(use_set, set_difference(out_set, def_set))
    # for node in result:
    #     if node in all_defs:
    #         continue
    return result


def print_table(index, new_result, iteration=0):
    table = [f"\nRDA: iteration {iteration}\n"]
    for key in new_result:
        try:
            table.append(str(read_index(index, key)[0][0] + 1) + "\n")
        except:
            table.append(str(key) + "\n")
        for key2 in new_result[key]:
            table.append(f"\t{key2}:")
            for entry in new_result[key][key2]:
                table.append(str(entry) + ",")
            table.append("\n")
    logger.debug("".join(table))


def start_rda(index, rda_table, graph, pre_solve=False):
    if debug and not pre_solve:
        print_table(index, rda_table)
    remove_set = []
    # "method_return", "class_return"
    if pre_solve:
        remove_set = ["method_call", "method_return", "class_return", "constructor_call"]
    graph = copy.deepcopy(graph)
    remove_edges = []
    if remove_set:
        for edge in graph.edges:
            if "label" in graph.edges[edge] and graph.edges[edge]["label"] in remove_set:
                remove_edges.append(edge)
        graph.remove_edges_from(remove_edges)

    cfg = graph
    nodes = graph.nodes
    old_result = {}
    iteration = 0
    for node in nodes:
        old_result[node] = {
            "IN": set(),
            "OUT": set()
        }
    # for node in nodes:
    #     connected_derivers = [t for (s, t) in cfg.out_edges(node)]
    #     old_result[node]["OUT"] = compute_out(old_result, connected_derivers, rda_table[node]["def"] if node in rda_table else [])
    new_result = copy.deepcopy(old_result)
    while True:
        iteration = iteration + 1
        iteration_returners = {}
        for node in nodes:
            connected_feeders = []
            returning_feeders = []
            for (s, t) in cfg.in_edges(node):
                if "label" in cfg.edges[s, t, 0] and cfg.edges[s, t, 0]["label"] in ["method_return", "class_return"]:
                    returning_feeders.append(s)
                else:
                    connected_feeders.append(s)
            feeder_union = set()
            returner_union = set()
            for feeder in connected_feeders:
                feeder_union = feeder_union.union(old_result[feeder]["OUT"])
            for returner in returning_feeders:
                returner_union = returner_union.union(old_result[returner]["OUT"])
            for (s, t) in cfg.out_edges(node):
                if s!=t:
                    iteration_returners[t] = iteration_returners[t].union(returner_union) if t in iteration_returners else returner_union
            new_result[node]["IN"] = feeder_union
            def_info = rda_table[node]["def"] if node in rda_table else set()
            names_defined = [defined.name for defined in def_info]
            # if defined.param_declaration is None
            selective_difference = set()
            for already_defined in feeder_union:
                # if already_defined.immutable_declaration:
                #     selective_difference.add(already_defined)
                if already_defined.name in names_defined:
                    continue
                else:
                    selective_difference.add(already_defined)
            new_result[node]["OUT"] = selective_difference.union(def_info)
        for node in iteration_returners:
            def_info = rda_table[node]["def"] if node in rda_table else set()
            names_defined = [defined.name for defined in def_info]
            selective_difference = set()
            for already_defined in iteration_returners[node]:
                # if already_defined.immutable_declaration:
                #     selective_difference.add(already_defined)
                if already_defined.name in names_defined:
                    continue
                else:
                    selective_difference.add(already_defined)
            new_result[node]["OUT"] = new_result[node]["OUT"].union(selective_difference)
        ddiff = DeepDiff(
            old_result,
            new_result,
            ignore_order=True,
        )
        if ddiff == {}:
            # for node in nodes:
            #     all_defs = []
            #     connected_feeders = [t for (s, t) in cfg.in_edges(node)]
            #     for feeder in connected_feeders:
            #         defs = rda_table[feeder]["def"] if feeder in rda_table else []
            #         all_defs += defs
            #     for outgoing in new_result[node]["IN"]:
            #         if outgoing in all_defs:
            #             continue
            #     connected_derivers = [t for (s, t) in cfg.out_edges(node)]
            #     for deriver in connected_derivers:
            #         for incoming in new_result[deriver]["IN"]:
            #             if incoming in rda_table[node]["def"]:
            #                 continue
            if debug and not pre_solve:
                logger.info("RDA: Completed in iteration {}", iteration)
                print_table(index, new_result, iteration)
            break
        old_result = copy.deepcopy(new_result)
    return new_result


def add_edge(final_graph, a, b, attrib=None, pre_solve=False):
    if (pre_solve and attrib) or not final_graph.has_edge(a, b):
        final_graph.add_edge(a, b)
        if attrib is not None:
            nx.set_edge_attributes(final_graph, {(a, b, 0): attrib})


def get_required_edges_from_def_to_use(index, cfg, rda_solution, rda_table, graph_nodes, all_classes,
                                       additional_edges=None, processed_edges=None, pre_solve=False, properties=None):
    if additional_edges is None:
        additional_edges = []
    if processed_edges is None:
        processed_edges = []
    final_graph = copy.deepcopy(cfg)
    # twin_edges = [
    #     "constructor_call", "method_call"
    # ]
    # twins = []
    # retain_edges = [
    #     "constructor_call", "class_return", "method_call", "method_return"
    # ]
    # retains = []
    # additional_edges = []
    # for edge in list(final_graph.edges()):
    #     edge_data = final_graph.get_edge_data(*edge)[0]
    #     if edge_data["label"] in retain_edges:
    #         retains.append(edge)
    #     # if edge_data["label"] in twin_edges:
    #     #     twins.append(edge)
    final_graph.remove_edges_from(list(final_graph.edges()))
    for node in graph_nodes:
        rda_info = rda_table[node] if node in rda_table else None
        if rda_info is not None:
            use_info = rda_info["use"]
            # def_info = rda_info["def"]
        else:
            use_info = set()
            # def_info = set()
        # names_used = [used.name for used in use_info]
        for used in use_info:
            for available_def in rda_solution[node]["IN"]:
                if available_def.name == used.name:
                    # if available_def not in rda_solution[node]["OUT"]:
                    #     for line_def in rda_table[node]["def"]:
                    #         if line_def.name == available_def.name:
                    #             if line_def.declaration:
                    #                 available_def = line_def
                    # if available_def in rda_table[available_def.line]["def"]:
                    if scope_check(available_def.scope, used.scope):
                        # if available_def.declaration:
                        # if available_def.immutable_declaration:
                        #     for other_def in rda_solution[node]["IN"]:
                        #         if other_def.name == available_def.name and other_def.param_declaration:
                        #             if not used.core.startswith("this"):
                        #                 available_def = other_def
                        #             break
                        add_edge(final_graph, available_def.line, node, {'used_def': used.name}, pre_solve)
                        used.satisfied = True
                    # else:
                    #         add_edge(final_graph, available_def.line, node, {'used_def': used.name}, pre_solve)
                    #         used.satisfied = True
                elif "." in used.name or "." in available_def.name:
                    if len(used.name.split(".")) < len(available_def.name.split(".")):
                        smaller = used
                        bigger = available_def
                    else:
                        smaller = available_def
                        bigger = used
                    if bigger.name.strip().startswith(smaller.name.strip()):
                        if not bigger.name.strip().replace(smaller.name.strip(), "").strip().startswith("."):
                            continue

                        # add_edge(final_graph, bigger.line or node, smaller.line or node)
                        add_edge(final_graph, available_def.line or node, used.line or node, {'used_def': used.name},
                                 pre_solve)
                        used.satisfied = True
            if not used.satisfied:
                if used.core in all_classes:
                    static_derive_class = all_classes[used.core]
                    fields = recursively_get_children_of_types(static_derive_class, "field_declaration")
                    for field in fields:
                        field_pieces = st(field).split()
                        if "static" in field_pieces:
                            if used.name.replace(used.core + ".", "") in field_pieces:
                                processed_edges.append((get_index(field, index), node))
                else:
                    if not pre_solve and properties.get("last_use", False):
                        for i in rda_table:
                            for entry in rda_table[i]["use"]:
                                if entry.name == used.name and i != node:
                                    # TODO: THIS IS AN AP - Make sure edges added here are right
                                    if i in cfg and node in cfg:
                                        # TODO: check missing i and handle it
                                        if not nx.has_path(cfg, i, node):
                                            continue
                                        add_edge(final_graph, i, node, {'color': 'green'})
                                    # used.satisfied = True
                            #         break
                            # if used.satisfied:
                            #     break
        if properties.get("last_def", False):
            for available_def in rda_solution[node]["IN"] - rda_solution[node]["OUT"]:
                ignore_nodes = ['for_statement', 'while_statement', 'if_statement', 'switch_statement']
                if read_index(index, node)[-1] in ignore_nodes or read_index(index, available_def.line)[
                    -1] in ignore_nodes:
                    continue
                add_edge(final_graph, available_def.line, node, {'color': 'orange'})
        # for definition in def_info:
        #     for available_def in rda_solution[node]["IN"]:
        #         if "." in definition.name or "." in available_def.name:
        #             if len(definition.name.split(".")) < len(available_def.name.split(".")):
        #                 smaller = definition
        #                 bigger = available_def
        #             else:
        #                 smaller = available_def
        #                 bigger = definition
        #             if bigger.name.strip().startswith(smaller.name.strip()):
        #                 if not bigger.name.strip().replace(smaller.name.strip(), "").strip().startswith("."):
        #                     continue
        #                 # add_edge(final_graph, bigger.line or node, smaller.line or node)
        #                 add_edge(final_graph, smaller.line or node, bigger.line or node)
    if not pre_solve:
        for edge in additional_edges:
            if debug:
                add_edge(final_graph, *edge, {'color': 'yellow'})
            else:
                add_edge(final_graph, *edge)
        for edge in processed_edges:
            add_edge(final_graph, *edge)
    # for edge in twins:
    #     continue
    #     add_edge(final_graph, edge[0], edge[1])
    #     add_edge(final_graph, edge[1], edge[0])
    return final_graph


def rda_cfg_map(rda_solution, CFG_results):
    graph = CFG_results.graph
    attrs = {}
    for edge in list(graph.edges):
        out_set = rda_solution[edge[0]]["OUT"]
        in_set = rda_solution[edge[1]]["IN"]
        intersection = set_intersection(out_set, in_set)
        data = graph.get_edge_data(*edge)
        if intersection:
            data['label'] = ",".join([str(intr) for intr in intersection])
        else:
            graph.remove_edge(*edge)
            # logger.warning("Unable to remap edge {}", edge)
        attrs[edge] = data
    nx.set_edge_attributes(graph, attrs)
    return graph


def node_has_parent(nvalue, parent_class):
    if nvalue == parent_class:
        return True
    while nvalue.parent:
        if nvalue.parent == parent_class:
            return True
        else:
            nvalue = nvalue.parent
    return False


def call_variable(call_variable_map, node, edge_for, check_virtual=True):
    for virtual_var, actual_var in call_variable_map[node]:
        if check_virtual:
            if st(virtual_var) == edge_for:
                return actual_var
        else:
            if st(actual_var) == edge_for:
                return virtual_var
    return None


def dfg_java(prop, CFG_results):
    # global properties
    properties = prop
    parser = CFG_results.parser
    index = parser.index
    tree = parser.tree

    call_variable_map = {}
    handled_cases = []

    # if_statement = ["if_statement", "else"]
    # for_statement = ["for_statement"]
    # enhanced_for_statement = ["for_each_statement"]
    # while_statement = ["while_statement"]

    additional_edges = []
    processed_edges = []
    cfg_graph = copy.deepcopy(CFG_results.graph)

    # Temperory edge adding logic when CFG is broken
    # 41 -> 57
    # 57 -> 27
    # add_edge(cfg_graph, 41, 57, {'controlflow_type': 'method_return', 'edge_type': 'CFG_edge', 'label': 'method_return', 'color': 'blue'})
    # add_edge(cfg_graph, 57, 27, {'controlflow_type': 'method_call', 'edge_type': 'CFG_edge', 'label': 'method_call', 'color': 'blue'})

    # global node_list
    node_list = CFG_results.node_list

    all_classes = {
        st(node_list[read_index(index, x)].child_by_field_name("name")): node_list[read_index(index, x)]
        for x, y in cfg_graph.nodes(data=True) if 'type_label' in y and y['type_label'] == "class_declaration"
    }
    uninitiated_classes = [x for x, y in cfg_graph.nodes(data=True) if 'type_label' in y and
                           y['type_label'] == "class_declaration" and cfg_graph.in_degree(x) == 0]

    start_rda_init_time = time.time()
    for edge in list(cfg_graph.edges()):
        edge_data = cfg_graph.get_edge_data(*edge)[0]
        # If there is object creation
        if edge_data["label"] == "class_return":
            # call_statement = node_list[read_index(index, edge[1])]
            # call_node = recursively_get_children_of_types(call_statement, method_invocation)[0]
            last_statement = node_list[read_index(index, edge[0])]
            class_node = return_first_parent_of_types(last_statement, "class_declaration")
            class_name = st(class_node.child_by_field_name("name"))
            class_methods = recursively_get_children_of_types(class_node, method_declaration)
            class_id = get_index(class_node, index)
            # find all simple paths between class_id and edge[0]
            all_paths = nx.all_simple_paths(cfg_graph, source=class_id, target=edge[0])
            # only_path = nx.shortest_path(cfg_graph, source=class_id, target=edge[0])
            unique_constructor_children = set()
            for path in all_paths:
                for node in path:
                    unique_constructor_children.add(node)
            if len(unique_constructor_children) == 0:
                unique_constructor_children.add(edge[0])
            # unique_constructor_children = set(only_path)
            # for nkey, nvalue in node_list.items():
            #     if node_has_parent(nvalue, class_node):
            #         unique_constructor_children.add(index[nkey])
            for class_method in class_methods:
                class_method_name = st(class_method.child_by_field_name("name"))
                if class_method_name != class_name:
                    add_edge(cfg_graph, edge[0], get_index(class_method, index))

            for n_id in unique_constructor_children:
                additional_edges.append((n_id, edge[1]))
        # Handle method_call edges and all constructor calls if there are parameters
        elif edge_data["label"] in ["constructor_call", "method_call"]:
            call_statement = node_list[read_index(index, edge[0])]
            if call_statement.type in method_invocation:
                call_node = call_statement
            else:
                marked_index = int(edge_data["controlflow_type"].rsplit("|",1)[-1])
                call_nodes = recursively_get_children_of_types(call_statement, method_invocation)
                for call_node in call_nodes:
                    if marked_index == get_index(call_node, index):
                        break
            method = node_list[read_index(index, edge[1])]
            if method.type == "class_declaration":
                class_node = method
                constructors = recursively_get_children_of_types(class_node, "constructor_declaration")
                if constructors:
                    if "target_constructor" in edge_data:
                        target_constructor = edge_data["target_constructor"]
                        for constructor in constructors:
                            if get_index(constructor, index) == target_constructor:
                                method = constructor
                                break
                    else:
                        continue
                        # TODO if this else is hit its an AP
                        # method = constructors[0]
                else:
                    continue
            actual_variables = call_node.child_by_field_name("arguments").named_children
            virtual_variables = []
            for node in method.named_children:
                if "formal_parameter" in node.type:
                    virtual_variables = recursively_get_children_of_types(node, variable_type, index=parser.index, stop_types=statement_types["statement_holders"])
            if actual_variables:
                # TODO: Handle var_args Type... IterableObject
                var_args = False
                if var_args or len(actual_variables) == len(virtual_variables):
                    for actual_var, virtual_var in zip(actual_variables, virtual_variables):
                        method_statement_id = edge[1]
                        if method_statement_id not in call_variable_map:
                            call_variable_map[method_statement_id] = []
                        call_variable_map[method_statement_id].append((virtual_var, actual_var))
                    processed_edges.append(edge)
                else:
                    logger.error("Number of actual and virtual variables do not match")
                    if not var_args:
                        pass
                        # assert len(actual_variables) == len(virtual_variables)
        elif edge_data["label"] == "method_return":
            return_statement = node_list[read_index(index, edge[0])]
            if return_statement.type == "return_statement" and return_statement.named_children:
                processed_edges.append(edge)
        elif edge_data["label"] == "lambda_invocation":
            processed_edges.append(edge)

    for class_id in uninitiated_classes:
        class_node = node_list[read_index(index, class_id)]
        class_name = st(class_node.child_by_field_name("name"))
        class_methods = recursively_get_children_of_types(class_node, method_declaration)

        descendants = [x for x in nx.descendants(cfg_graph, class_id) if
                       cfg_graph.out_degree(x) == 0 and cfg_graph.in_degree(x) == 1]

        for class_method in class_methods:
            class_method_name = st(class_method.child_by_field_name("name"))
            if class_method_name != class_name:
                for descendant in descendants:
                    add_edge(cfg_graph, descendant, get_index(class_method, index))

    rda_table = {}
    for root_node in traverse_tree(tree, variable_type):
        if not root_node.is_named:
            if root_node.type == "while" and root_node.parent.type == "do_statement":
                root_node = root_node.next_named_sibling
                parent_id = get_index(root_node, index)
                identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index, stop_types=statement_types["statement_holders"],
                                                                     check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    # if identifier.parent.type == "method_invocation":
                    #     continue
                    add_entry(parser, rda_table, parent_id, used=identifier)
            else:
                continue
        if root_node.type == "catch_clause":
            parent_id = get_index(root_node, index)
            if len(root_node.named_children) > 1:
                identifiers_used = recursively_get_children_of_types(root_node.named_children[-2], variable_type,
                                                                     index=parser.index,
                                                                     check_list=parser.symbol_table["scope_map"])
                add_entry(parser, rda_table, parent_id, defined=identifiers_used[-1])
        if root_node.type == "return_statement":
            parent_id = get_index(root_node, index)
            identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                if identifier.parent.type == "method_invocation":
                    continue
                add_entry(parser, rda_table, parent_id, used=identifier)
        if root_node.type in def_statement:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_statement is None:
                continue
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            # print(parent_id)
            # print(st(parent_statement))
            if len(root_node.named_children) == 2:
                name = root_node.named_children[0]
                value = root_node.named_children[1]
            else:
                name = root_node.named_children[0]
                value = None
            add_entry(parser, rda_table, parent_id, defined=name, declaration=True)
            if value is not None:
                if value.type == "identifier":
                    identifiers_used = [value]
                else:
                    identifiers_used = recursively_get_children_of_types(value, variable_type, index=parser.index, stop_types=statement_types["statement_holders"],
                                                                         check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    if identifier.parent.type == "method_invocation":
                        continue
                    add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in assignment:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            # print(parent_id)
            # print(st(parent_statement))
            left_nodes = root_node.child_by_field_name("left")
            right_node = root_node.child_by_field_name("right")
            operator_text = (
                root_node.text.split(left_nodes.text, 1)[-1]
                .rsplit(right_node.text, 1)[0]
                .strip()
                .decode()
            )
            right_nodes = [right_node]
            # JRef
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            # operator = root_node.child()
            # operator_text = operator.text.decode('utf-8')
            if operator_text != "=":
                right_nodes.append(left_nodes)
            # TODO triage
            add_entry(parser, rda_table, parent_id, defined=left_nodes)
            for node in right_nodes:
                if node.type in variable_type:
                    add_entry(parser, rda_table, parent_id, used=node)
                else:
                    identifiers_used = recursively_get_children_of_types(node, variable_type, index=parser.index, stop_types=statement_types["statement_holders"],
                                                                         check_list=parser.symbol_table["scope_map"])
                    for identifier in identifiers_used:
                        if identifier.parent.type == "method_invocation":
                            continue
                        add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in increment_statement:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            # print(parent_id)
            # print(st(parent_statement))
            if root_node.type in variable_type:
                add_entry(parser, rda_table, parent_id, used=root_node)
            else:
                identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index, stop_types= statement_types["statement_holders"],
                                                                     check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    if identifier.parent.type == "method_invocation":
                        continue
                    add_entry(parser, rda_table, parent_id, used=identifier)
                    add_entry(parser, rda_table, parent_id, defined=identifier)
        elif root_node.type in method_calls:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement is None:
                    continue
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            add_entry(parser, rda_table, parent_id, used=root_node.children[0], method_call=True)
            for node in root_node.children[1:]:
                if node.type in variable_type:
                    if return_first_parent_of_types(node, 'method_invocation') is None:
                        add_entry(parser, rda_table, parent_id, used=node)
                else:
                    identifiers_used = recursively_get_children_of_types(node, variable_type, index=parser.index, stop_types=statement_types["statement_holders"],
                                                                         check_list=parser.symbol_table["scope_map"])
                    for identifier in identifiers_used:
                        if identifier.parent.type == "method_invocation":
                            continue
                        add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in method_declaration:
            parent_id = get_index(root_node, index)
            for child in root_node.named_children:
                # if root_node.type == "constructor_declaration":
                #     if child.type == "identifier":

                if child.type == "formal_parameters":
                    for parameter in child.named_children:
                        if parameter.type == "formal_parameter":
                            identifier = parameter.child_by_field_name("name")
                            add_entry(parser, rda_table, parent_id, defined=identifier)
                    break
        elif root_node.type in switch_type:
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            if parent_statement.type == 'switch_expression':
                add_entry(parser, rda_table, parent_id, used=root_node.children[0])
            elif parent_statement.type == 'switch_statement':
                case_node = parent_statement.child_by_field_name("value")
                add_entry(parser, rda_table, parent_id, used=case_node)
        elif root_node.type == "throw_statement":
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            identifiers_used = recursively_get_children_of_types(parent_statement, variable_type, index=parser.index, stop_types=statement_types["statement_holders"],
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                if identifier.parent.type == "method_invocation":
                    continue
                add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type == "enhanced_for_statement":
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            for i, child in enumerate(parent_statement.children):
                if child.parent.field_name_for_child(i) == "dimensions":
                    defined_node = child
                    add_entry(parser, rda_table, parent_id, defined=defined_node, declaration=True)
                elif child.parent.field_name_for_child(i) == "value":
                    used_node = child
                    add_entry(parser, rda_table, parent_id, used=used_node)
        elif root_node.type == "for_statement":
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            for i, child in enumerate(parent_statement.children):
                if child.parent.field_name_for_child(i) == "dimensions":
                    defined_node = child
                    add_entry(parser, rda_table, parent_id, defined=defined_node, declaration=True)
                elif child.parent.field_name_for_child(i) == "value":
                    used_node = child
                    add_entry(parser, rda_table, parent_id, used=used_node)
        elif root_node.type == "object_creation_expression":
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            identifiers_used = recursively_get_children_of_types(root_node, variable_type,
                                                                 index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                if identifier.parent.type == "method_invocation":
                    continue
                add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type == "lambda_expression":
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            defined_nodes = parent_statement.child_by_field_name("parameters")
            if defined_nodes:
                for defined_node in defined_nodes.named_children:
                    add_entry(parser, rda_table, parent_id, defined=defined_node, declaration=True)
        else:
            # THIS ELSE IS VERY INEFFICIENT
            # Essentially considers all variable references as use
            # if return_first_parent_of_types(root_node, handled_types) is None:
            parent_statement = return_first_parent_of_types(
                root_node,
                # statement_types["node_list_type"],
                # [x for x in statement_types["node_list_type"] if x not in block_types],
                statement_types["non_control_statement"] + statement_types["control_statement"],
                stop_types=statement_types["statement_holders"] + handled_types,
                remap=False
            )
            parent_id = get_index(parent_statement, index)
            if parent_id is None:
                continue
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in inner_types:
                    parent_statement = return_first_parent_of_types(parent_statement, statement_types["node_list_type"])
                    parent_id = get_index(parent_statement, index)
                elif parent_statement.type in handled_cases:
                    continue
                else:
                    raise NotImplementedError
            # print(root_node.type)
            # print(parent_statement.type)

            # print("...")

            # print(parent_id)
            # print(st(parent_statement))

            # if parent_statement and parent_statement.type in block_types:
            #     continue

            # if root_node.type in variable_type:
            #     add_entry(parser, rda_table, parent_id, used=root_node)
            # else:
            identifiers_used = recursively_get_children_of_types(root_node, variable_type, stop_types=statement_types[
                                                                                                          "statement_holders"] + handled_types,
                                                                 index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                if identifier.parent.type == "method_invocation":
                    continue
                add_entry(parser, rda_table, parent_id, used=identifier)
    end_rda_init_time = time.time()
    end_rda_presolve_time = 0
    start_rda_presolve_time = 0
    end_alias_analysis_time = 0
    start_alias_analysis_time = 0
    if properties.get("alex_algo", True):
        start_rda_presolve_time = time.time()
        rda_solution = start_rda(index, rda_table, cfg_graph, pre_solve=True)
        # pp.pprint(rda_solution)
        final_graph = get_required_edges_from_def_to_use(index, cfg_graph, rda_solution, rda_table,
                                                         cfg_graph.nodes, all_classes, additional_edges, processed_edges,
                                                         pre_solve=True, properties=properties)
        used_classes_or_methods = set()
        for edge in cfg_graph.edges:
            if "label" in cfg_graph.edges[edge] and cfg_graph.edges[edge]["label"] in ["method_call", "constructor_call"]:
                final_graph.nodes[edge[0]][cfg_graph.edges[edge]["label"]] = edge[1]
                used_classes_or_methods.add(edge[1])
        end_rda_presolve_time = time.time()
        start_alias_analysis_time = time.time()
        al_analysis = set()
        proc_analysis = set()

        for node in final_graph.nodes:
            if node not in used_classes_or_methods:
                continue
            if final_graph.nodes[node]["type_label"] == "method_declaration":
                if "main" in final_graph.nodes[node]["label"]:
                    continue
                handled_node_leaf_pair = set()
                for leaf in nx.algorithms.dag.descendants(final_graph, node):
                    if len([x for x in final_graph.edges(leaf) if x[0] != x[1]]) == 0:
                        # print("Leaf: ", leaf)
                        # final_graph.add_edge(leaf, "exit")
                        # used_names = [n.name for n in rda_table[leaf]['def']]
                        # if len(used_names) != 1:
                        #     logger.error("Multiple used names: %s", used_names)
                        # get shortest path from node to leaf
                        paths = nx.all_simple_paths(final_graph, node, leaf)
                        # print("All_Simple: ", node, leaf)
                        for path in paths:
                            # get first edge in path
                            first_edge = (path[0], path[1], 0)
                            # get last edge in path
                            last_edge = (path[-2], path[-1], 0)
                            # get data for edge
                            edge_for = final_graph.edges[first_edge]["used_def"]
                            # get data for last edge
                            edge_for_last = final_graph.edges[last_edge]["used_def"]
                            defined_cores = [d.core for d in rda_table[leaf]["def"]]
                            call_node = None
                            if edge_for_last not in defined_cores:
                                # "method_return", "class_return"
                                if any([x in final_graph.nodes[leaf] for x in ["method_call", "constructor_call"]]):
                                    if "method_call" in final_graph.nodes[leaf]:
                                        call_node = final_graph.nodes[leaf]["method_call"]
                                    elif "constructor_call" in final_graph.nodes[leaf]:
                                        call_node = final_graph.nodes[leaf]["constructor_call"]
                            node_leaf_key = str(leaf) + "_" + str(edge_for)
                            if node_leaf_key in handled_node_leaf_pair:
                                continue
                            handled_node_leaf_pair.add(node_leaf_key)
                            if node in call_variable_map:
                                if call_node:
                                    proc_analysis.add((node, leaf, edge_for, call_node))
                                # al_analysis.append((node, leaf, edge_for, call_node))
                                if edge_for_last not in defined_cores:
                                    n = -1
                                    while edge_for_last not in defined_cores:
                                        n = n - 1
                                        if abs(n-1) > len(path):
                                            break
                                        last_edge = (path[n - 1], path[n], 0)
                                        edge_for_last = final_graph.edges[last_edge]["used_def"]
                                        defined_cores = [d.core for d in rda_table[path[n]]["def"]]
                                        if edge_for_last in defined_cores:
                                            al_analysis.add((node, path[n], edge_for, None))
                                else:
                                    al_analysis.add((node, leaf, edge_for, call_node))

        while proc_analysis:
            temp_proc_analysis = set()
            for node, leaf, edge_for, call_node in proc_analysis:
                completed_analysis = set()
                for al_entry in al_analysis:
                    if al_entry[0] == call_node:
                        mapped_node = call_variable(call_variable_map, call_node, edge_for, check_virtual=False)
                        if al_entry[2] != st(mapped_node):
                            continue
                        leaf = al_entry[1]
                        completed_analysis.add((node, leaf, edge_for, al_entry[3]))
                        if al_entry[3] is not None:
                            temp_proc_analysis.add((node, leaf, edge_for, al_entry[3]))
                al_analysis.update(completed_analysis)
            proc_analysis = temp_proc_analysis

        for node, leaf, edge_for, call_node in al_analysis:
            mapped_node = call_variable(call_variable_map, node, edge_for)
            if mapped_node:
                if mapped_node.type == "method_invocation":
                    # if a reference is passed back it is ignored in alias analysis
                    # to handle turn this into a while loop and trace the return chain
                    #     track_leaf = leaf
                    #     leaf_node = node_list[read_index(index, track_leaf)]
                    #     if leaf_node.type == "return_statement":
                    continue
                # [(st(a),st(b)) for a,b in call_variable_map[100]]
                # print(edge_for, final_graph.nodes[leaf]["label"])
                set_add(rda_table[leaf]["def"],
                        Identifier(parser, mapped_node, leaf, full_ref=mapped_node,
                                   declaration=True,
                                   method_call=True))
        end_alias_analysis_time = time.time()
    # print_table(index, rda_table)
    start_rda_time = time.time()
    rda_solution = start_rda(index, rda_table, cfg_graph)
    final_graph = get_required_edges_from_def_to_use(index, cfg_graph, rda_solution, rda_table,
                                                     cfg_graph.nodes, all_classes, additional_edges, processed_edges,
                                                     pre_solve=False, properties=properties)
    end_rda_time = time.time()
    if debug:
        logger.warning("RDA init, presolve, alias, rda: {}, {}, {}, {}", end_rda_init_time - start_rda_init_time,
                         end_rda_presolve_time - start_rda_presolve_time, end_alias_analysis_time - start_alias_analysis_time,
                            end_rda_time - start_rda_time)
    debug_graph = rda_cfg_map(rda_solution, CFG_results)
    return final_graph, debug_graph, rda_table, rda_solution



from ..tree_parser.java_parser import JavaParser
from ..tree_parser.cs_parser import CSParser
from ..utils import preprocessor


class ParserDriver:
    """Driver class for the parser"""

    def __init__(self, src_language, src_code):
        """Initialize the driver. Preprocess the code before parsing"""
        self.src_language = src_language
        self.src_code = self.pre_process_src_code(src_language, src_code)

        self.parser_map = {
            "java": JavaParser,
            "cs": CSParser,
            # 'python': PythonParser
            # Add more languages here
        }
        self.parser = self.parser_map[self.src_language](self.src_language, self.src_code)
        self.root_node, self.tree = self.parser.parse()
        (
            self.all_tokens,
            self.label,
            self.method_map,
            self.method_calls,
            self.start_line,
            self.declaration,
            self.declaration_map,
            self.symbol_table,
        ) = self.create_all_tokens()


    def pre_process_src_code(self, src_language, src_code):
        """Pre-process the source code"""
        src_code = preprocessor.remove_empty_lines(src_code)
        src_code = preprocessor.remove_comments(src_language, src_code)
        return src_code

    # Needs to be modified for every language
    def create_all_tokens(self):
        """Return the variable list"""
        return self.parser.create_all_tokens(
            self.src_code,
            self.parser.root_node,
            self.parser.all_tokens,
            self.parser.label,
            self.parser.method_map,
            self.parser.method_calls,
            self.parser.start_line,
            self.parser.declaration,
            self.parser.declaration_map,
            self.parser.symbol_table,
        )



# This code is from https://github.com/microsoft/CodeBERT/blob/master/GraphCodeBERT/clonedetection/parser/utils.py

def tree_to_token_index(root_node):
    """Returns all tokens in the tree rooted at the given root node using index values"""
    if (
        len(root_node.children) == 0 or root_node.type == "string"
    ) and root_node.type != "comment":
        return [(root_node.start_point, root_node.end_point, root_node.type)]
    else:
        code_tokens = []
        for child in root_node.children:
            code_tokens += tree_to_token_index(child)
        return code_tokens


def tree_to_variable_index(root_node, index_to_code):
    """Returns all tokens in the tree rooted at the given root node"""
    if (
        len(root_node.children) == 0 or root_node.type == "string"
    ) and root_node.type != "comment":
        index = (root_node.start_point, root_node.end_point, root_node.type)
        _, code = index_to_code[index]
        if root_node.type != code:
            return [(root_node.start_point, root_node.end_point, root_node.type)]
        else:
            return []
    else:
        code_tokens = []
        for child in root_node.children:
            code_tokens += tree_to_variable_index(child, index_to_code)
        return code_tokens


def index_to_code_token(index, code):
    """Returns code value given the indexes of a particular token"""
    start_point = index[0]
    end_point = index[1]
    if start_point[0] == end_point[0]:
        s = code[start_point[0]][start_point[1] : end_point[1]]
    else:
        s = ""
        s += code[start_point[0]][start_point[1] :]
        for i in range(start_point[0] + 1, end_point[0]):
            s += code[i]
        s += code[end_point[0]][: end_point[1]]
    return s
